__author__ = 'aarongary'

import json
import sys
import os
import errno
import pandas as pd
import networkx as nx
import ndex2.client as nc
import io
import decimal
import numpy as np
import math
import json
import ijson
import requests
import base64
import nicecxModel.NiceCXStreamer as ncs
from nicecxModel.metadata.MetaDataElement import MetaDataElement
from nicecxModel.cx.aspects.NameSpaces import NameSpaces
from nicecxModel.cx.aspects.NodesElement import NodesElement
from nicecxModel.cx.aspects.EdgesElement import EdgesElement
from nicecxModel.cx.aspects.NodeAttributesElement import NodeAttributesElement
from nicecxModel.cx.aspects.EdgeAttributesElement import EdgeAttributesElement
from nicecxModel.cx.aspects.NetworkAttributesElement import NetworkAttributesElement
from nicecxModel.cx.aspects.SupportElement import SupportElement
from nicecxModel.cx.aspects.CitationElement import CitationElement
from nicecxModel.cx.aspects.AspectElement import AspectElement
from nicecxModel.cx import CX_CONSTANTS
from nicecxModel.cx.aspects import ATTRIBUTE_DATA_TYPE
from nicecxModel.cx import known_aspects, known_aspects_min

if sys.version_info.major == 3:
    from urllib.request import urlopen, Request, HTTPBasicAuthHandler, HTTPPasswordMgrWithDefaultRealm, \
        build_opener, install_opener, HTTPError, URLError
else:
    from urllib2 import urlopen, Request, HTTPBasicAuthHandler, HTTPPasswordMgrWithDefaultRealm, \
        build_opener, install_opener, HTTPError, URLError

class NiceCXNetwork(object):
    def __init__(self, cx=None, server=None, username=None, password=None, uuid=None, networkx_G=None, pandas_df=None, filename=None, data=None, **attr):
        self.metadata = {}
        self.namespaces = NameSpaces()
        self.nodes = {}
        self.node_int_id_generator = set([])
        self.edge_int_id_generator = set([])
        self.node_id_lookup = []
        self.edges = {}
        self.citations = {}
        self.nodeCitations = {}
        self.edgeCitations = {}
        self.edgeSupports = {}
        self.nodeSupports = {}
        self.supports = {}
        self.nodeAttributes = {}
        self.edgeAttributes = {}
        self.edgeAttributeHeader = set([])
        self.nodeAttributeHeader = set([])
        self.networkAttributes = []
        self.nodeAssociatedAspects = {}
        self.edgeAssociatedAspects = {}
        self.opaqueAspects = {}
        self.provenance = None
        self.missingNodes = {}

        if cx:
            self.create_from_cx(cx)
        elif networkx_G:
            self.create_from_networkx(networkx_G)
        elif pandas_df is not None:
            self.create_from_pandas(pandas_df)
        elif filename is not None:
            if os.path.isfile(filename):
                with open(filename, 'rU') as file_cx:
                    #====================================
                    # BUILD NICECX FROM FILE
                    #====================================
                    self.create_from_cx(json.load(file_cx))
            else:
                raise Exception('Input provided is not a valid file')
        else:
            if server and uuid:
                self.create_from_server(server, username, password, uuid)

    def addNode(self, node):
        if type(node) is NodesElement:
            self.nodes[node.getId()] = node

            if type(node.getId()) is str:
                self.node_int_id_generator.add(node.getId())

            if self.missingNodes.get(node.getId()) is not None:
                self.missingNodes.pop(node.getId(), None)
        else:
            raise Exception('Provided input was not of type NodesElement.')

    def addEdge(self, edge):
        if type(edge) is EdgesElement:
            if edge.getId() < 0:
                edge.setId(len(self.edges.keys()))
            self.edges[edge.getId()] = edge

            if self.nodes.get(edge.getSource()) is None:
                self.missingNodes[edge.getSource()] = 1

            if self.nodes.get(edge.getTarget()) is None:
                self.missingNodes[edge.getTarget()] = 1
        else:
            raise Exception('Provided input was not of type EdgesElement.')

    def addNetworkAttribute(self, network_attribute):
        if type(network_attribute) is NetworkAttributesElement:
            self.networkAttributes.append(network_attribute)
        else:
            raise Exception('Provided input was not of type NetworkAttributesElement.')

    def addNodeAttribute(self, node_attribute, i=None):
        if type(node_attribute) is NodeAttributesElement:
            self.nodeAttributeHeader.add(node_attribute.getName())
            nodeAttrs = self.nodeAttributes.get(node_attribute.getPropertyOf())
            if nodeAttrs is None:
                nodeAttrs = []
                self.nodeAttributes[node_attribute.getPropertyOf()] = nodeAttrs

            nodeAttrs.append(node_attribute)
        else:
            raise Exception('Provided input was not of type NodeAttributesElement.')

    def addEdgeAttribute(self, edge_attribute, i=None):
        if type(edge_attribute) is EdgeAttributesElement:
            self.edgeAttributeHeader.add(edge_attribute.getName())
            edgeAttrs = self.edgeAttributes.get(edge_attribute.getPropertyOf())
            if edgeAttrs is None:
                    edgeAttrs = []
                    self.edgeAttributes[edge_attribute.getPropertyOf()] = edgeAttrs

            edgeAttrs.append(edge_attribute)
        else:
            raise Exception('Provided input was not of type EdgeAttributesElement.')

    def addSupport(self, support_element):
        if type(support_element) is SupportElement:
            self.supports[support_element.getId()] = support_element
        else:
            raise Exception('Provided input was not of type SupportElement.')

    def addCitation(self, citation_element):
        if type(citation_element) is CitationElement:
            self.citations[citation_element.getId()] = citation_element
        else:
            raise Exception('Provided input was not of type CitationElement.')

    def addNodeCitationsFromCX(self, node_citation_cx):
        self.buildManyToManyRelation('nodeCitations', node_citation_cx, 'citations')

    def addNodeCitations(self, node_id, citation_id):
        node_citation_element = {CX_CONSTANTS.PROPERTY_OF: [node_id], CX_CONSTANTS.CITATIONS: [citation_id]}
        self.buildManyToManyRelation('nodeCitations', node_citation_element, 'citations')

    def addEdgeCitationsFromCX(self, edge_citation_cx):
        self.buildManyToManyRelation('edgeCitations', edge_citation_cx, 'citations')

    def addEdgeCitations(self, edge_id, citation_id):
        edge_citation_element = {CX_CONSTANTS.PROPERTY_OF: [edge_id], CX_CONSTANTS.CITATIONS: [citation_id]}
        self.buildManyToManyRelation('edgeCitations', edge_citation_element, 'citations')

    def addEdgeSupports(self, edge_supports_element):
        self.buildManyToManyRelation('edgeSupports', edge_supports_element, 'supports')

    def buildManyToManyRelation(self, aspect_name, element, relation_name):
        if aspect_name == 'nodeCitations':
            aspect = self.nodeCitations
        elif aspect_name == 'edgeCitations':
            aspect = self.edgeCitations
        elif aspect_name == 'edgeSupports':
            aspect = self.edgeSupports
        else:
            raise Exception('Only nodeCitations, edgeCitations and edgeSupports are supported. ' + aspect_name + ' was supplied')

        for po in element.get(CX_CONSTANTS.PROPERTY_OF):
            po_id = aspect.get(po)
            if po_id is None:
                aspect[po] = element.get(relation_name)
            else:
                aspect[po] += element.get(relation_name)

    def addOpapqueAspect(self, opaque_element):
        if type(opaque_element) is AspectElement:
            aspectElmts = self.opaqueAspects.get(opaque_element.getAspectName())
            if aspectElmts is None:
                aspectElmts = []
                self.opaqueAspects[opaque_element.getAspectName()] = aspectElmts

            aspectElmts.append(opaque_element.getAspectElement())
        else:
            raise Exception('Provided input was not of type AspectElement.')

    def addNodeAssociatedAspectElement(self, nodeId, elmt):
        self.addAssciatatedAspectElement(self.nodeAssociatedAspects, nodeId, elmt)

    def addEdgeAssociatedAspectElement(self, edgeId, elmt):
        self.addAssciatatedAspectElement(self.edgeAssociatedAspects, edgeId, elmt)

    def addAssciatatedAspectElement(self, table, id, elmt):
        aspectElements = table.get(elmt.getAspectName())
        if aspectElements is None:
            aspectElements = {}
            table.put(elmt.getAspectName(), aspectElements)

        elmts = aspectElements.get(id)

        if (elmts is None):
            elmts = []
            aspectElements.put(id, elmts)

        elmts.append(elmt)

    def setName(self, network_name):
        add_this_network_attribute = NetworkAttributesElement(name='name', values=network_name, type=ATTRIBUTE_DATA_TYPE.STRING)

        self.addNetworkAttribute(add_this_network_attribute)

    def getName(self):
        for net_a in self.networkAttributes:
            if net_a.getName() == 'name':
                return net_a.getValues()

        return None

    def getMetadata(self):
        return self.metadata

    def setMetadata(self, metadata_obj):
        if type(metadata_obj) is dict:
            self.metadata = metadata_obj
        else:
            raise Exception('Set metadata input was not of type <dict>')

    def addMetadata(self, md):
        if type(md) is MetaDataElement:
            #  TODO - alter metadata to match the element counts
            self.metadata[md.getName()] = md
        else:
            raise Exception('Provided input was not of type <MetaDataElement>')

    def addNameSpace(self, prefix, uri):
        self.namespaces[prefix] = uri

    def setNamespaces(self,ns ):
        self.namespaces = ns

    def getNamespaces(self,):
        return self.namespaces

    def getEdges (self):
        return self.edges

    def getNodes(self):
        return self.nodes

    def getOpaqueAspectTable(self):
        return self.opaqueAspects

    def getNetworkAttributes(self):
        return self.networkAttributes

    def getNodeAttributes(self):
        return self.nodeAttributes

    def getEdgeAttributes(self):
        return self.edgeAttributes

    def getNodeAssociatedAspects(self):
        return self.nodeAssociatedAspects

    def getEdgeAssociatedAspects(self):
        return self.edgeAssociatedAspects

    def getNodeAssociatedAspect(self, aspectName):
        return self.nodeAssociatedAspects.get(aspectName)

    def getEdgeAssociatedAspect(self, aspectName):
        return self.edgeAssociatedAspects.get(aspectName)

    def getProvenance(self):
        return self.provenance

    def getMissingNodes(self):
        return self.missingNodes

    def setProvenance(self, provenance):
        self.provenance = provenance

    def getEdgeCitations(self):
        return self.edgeCitations

    def getNodeCitations(self):
        return self.nodeCitations

    def apply_template(self, server, username, password, uuid):
        error_message = []
        if not server:
            error_message.append('server')
        if not uuid:
            error_message.append('uuid')

        if server and uuid:
            #===================
            # METADATA
            #===================
            available_aspects = []
            for ae in (o for o in self.streamAspect(uuid, 'metaData', server, username, password)):
                available_aspects.append(ae.get(CX_CONSTANTS.METADATA_NAME))

            #=======================
            # ADD VISUAL PROPERTIES
            #=======================
            for oa in available_aspects:
                if 'visualProperties' in oa:
                    objects = self.streamAspect(uuid, 'visualProperties', server, username, password)
                    obj_items = (o for o in objects)
                    for oa_item in obj_items:
                        aspect_element = AspectElement(oa_item, oa)
                        self.addOpapqueAspect(aspect_element)
                    vis_prop_size = len(self.opaqueAspects.get('visualProperties'))
                    mde = MetaDataElement(elementCount=vis_prop_size, version=1, consistencyGroup=1, name='visualProperties')
                    self.addMetadata(mde)


                if 'cyVisualProperties' in oa:
                    objects = self.streamAspect(uuid, 'cyVisualProperties', server, username, password)
                    obj_items = (o for o in objects)
                    for oa_item in obj_items:
                        aspect_element = AspectElement(oa_item, oa)
                        self.addOpapqueAspect(aspect_element)
                    vis_prop_size = len(self.opaqueAspects.get('cyVisualProperties'))
                    mde = MetaDataElement(elementCount=vis_prop_size, version=1, consistencyGroup=1, name='cyVisualProperties')
                    self.addMetadata(mde)
        else:
            raise Exception(', '.join(error_message) + 'not specified in apply_template')

    def create_from_pandas(self, df, source_field=None, target_field=None, source_node_attr=[], target_node_attr=[], edge_attr=[]):
        """
        Constructor that uses a pandas dataframe to build niceCX
        :param df: dataframe
        :type df: Pandas Dataframe
        :param headers:
        :type headers:
        :return: none
        :rtype: n/a
        """

        #====================================================
        # IF NODE FIELD NAME (SOURCE AND TARGET) IS PROVIDED
        # THEN USE THOSE FIELDS OTHERWISE USE INDEX 0 & 1
        #====================================================
        self.setName('Pandas Upload')
        self.add_metadata_stub('networkAttributes')
        count = 0
        if source_field and target_field:
            for index, row in df.iterrows():
                if count % 10000 == 0:
                    print count
                count += 1
                #=============
                # ADD NODES
                #=============
                self.addNode(NodesElement(id=row[source_field], node_name=row[source_field], node_represents=row[source_field]))
                self.addNode(NodesElement(id=row[target_field], node_name=row[target_field], node_represents=row[target_field]))

                #=============
                # ADD EDGES
                #=============
                self.addEdge(EdgesElement(id=index, edge_source=row[source_field], edge_target=row[target_field], edge_interaction='interacts-with'))

                #==============================
                # ADD SOURCE NODE ATTRIBUTES
                #==============================
                for sp in source_node_attr:
                    attr_type = None
                    if type(row[sp]) is float and math.isnan(row[sp]):
                        row[sp] = ''
                        attr_type = ATTRIBUTE_DATA_TYPE.FLOAT
                    elif type(row[sp]) is float and math.isinf(row[sp]):
                        row[sp] = 'Inf'
                        attr_type = ATTRIBUTE_DATA_TYPE.FLOAT
                    self.addNodeAttribute(NodeAttributesElement(property_of=row[source_field], name=sp, values=row[sp], type=attr_type))

                #==============================
                # ADD TARGET NODE ATTRIBUTES
                #==============================
                for tp in target_node_attr:
                    attr_type = None
                    if type(row[tp]) is float and math.isnan(row[tp]):
                        row[tp] = ''
                        attr_type = ATTRIBUTE_DATA_TYPE.FLOAT
                    elif type(row[tp]) is float and math.isinf(row[tp]):
                        row[tp] = 'Inf'
                        attr_type = ATTRIBUTE_DATA_TYPE.FLOAT
                    self.addNodeAttribute(NodeAttributesElement(property_of=row[target_field], name=tp, values=row[tp], type=attr_type))

                #==============================
                # ADD EDGE ATTRIBUTES
                #==============================
                for ep in edge_attr:
                    attr_type = None
                    if type(row[ep]) is float and math.isnan(row[ep]):
                        row[ep] = ''
                        attr_type = ATTRIBUTE_DATA_TYPE.FLOAT
                    elif type(row[ep]) is float and math.isinf(row[ep]):
                        row[ep] = 'INFINITY'
                        attr_type = ATTRIBUTE_DATA_TYPE.FLOAT

                    self.addEdgeAttribute(EdgeAttributesElement(property_of=index, name=ep, values=row[ep], type=attr_type))

        else:
            for index, row in df.iterrows():
                #=============
                # ADD NODES
                #=============
                self.addNode(NodesElement(id=row[0], node_name=row[0], node_represents=row[0]))
                self.addNode(NodesElement(id=row[1], node_name=row[1], node_represents=row[1]))

                #=============
                # ADD EDGES
                #=============
                if len(row) > 2:
                    self.addEdge(EdgesElement(id=index, edge_source=row[0], edge_target=row[1], edge_interaction=row[2]))
                else:
                    self.addEdge(EdgesElement(id=index, edge_source=row[0], edge_target=row[1], edge_interaction='interacts-with'))

        self.add_metadata_stub('nodes')
        self.add_metadata_stub('edges')
        if source_node_attr or target_node_attr:
            self.add_metadata_stub('nodeAttributes')
        if edge_attr:
            self.add_metadata_stub('edgeAttributes')

    def create_from_networkx(self, G):
        """
        Constructor that uses a networkx graph to build niceCX
        :param G: networkx graph
        :type G: networkx graph
        :return: none
        :rtype: none
        """
        self.setName('Networkx Upload')
        self.add_metadata_stub('networkAttributes')
        for n, d in G.nodes_iter(data=True):
            #=============
            # ADD NODES
            #=============
            self.addNode(NodesElement(id=n, node_name=n, node_represents=n))

            #======================
            # ADD NODE ATTRIBUTES
            #======================
            for k,v in d.items():
                self.addNodeAttribute(NodeAttributesElement(property_of=n, name=k, values=v))

        index = 0
        for u, v, d in G.edges_iter(data=True):
            #=============
            # ADD EDGES
            #=============
            self.addEdge(EdgesElement(id=index, edge_source=u, edge_target=v, edge_interaction=d.get('interaction')))

            #==============================
            # ADD EDGE ATTRIBUTES
            #==============================
            for k,v in d.items():
                self.addEdgeAttribute(EdgeAttributesElement(property_of=index, name=k, values=v))

            index += 1

        self.add_metadata_stub('nodes')
        self.add_metadata_stub('edges')
        if self.nodeAttributes:
            self.add_metadata_stub('nodeAttributes')
        if self.edgeAttributes:
            self.add_metadata_stub('edgeAttributes')
        #print json.dumps(self.to_json())

    def create_from_server(self, server, username, password, uuid):
        if server and uuid:
            #===================
            # METADATA
            #===================
            available_aspects = []
            for ae in (o for o in self.streamAspect(uuid, 'metaData', server, username, password)):
                available_aspects.append(ae.get(CX_CONSTANTS.METADATA_NAME))
                mde = MetaDataElement(json_obj=ae)
                self.addMetadata(mde)

            opaque_aspects = set(available_aspects).difference(known_aspects_min)

            #====================
            # NETWORK ATTRIBUTES
            #====================
            if 'networkAttributes' in available_aspects:
                objects = self.streamAspect(uuid, 'networkAttributes', server, username, password)
                obj_items = (o for o in objects)
                for network_item in obj_items:
                    add_this_network_attribute = NetworkAttributesElement(json_obj=network_item)

                    self.addNetworkAttribute(add_this_network_attribute)
                self.add_metadata_stub('networkAttributes')

            #===================
            # NODES
            #===================
            if 'nodes' in available_aspects:
                objects = self.streamAspect(uuid, 'nodes', server, username, password)
                obj_items = (o for o in objects)
                for node_item in obj_items:
                    add_this_node = NodesElement(json_obj=node_item)

                    self.addNode(add_this_node)
                self.add_metadata_stub('nodes')

            #===================
            # EDGES
            #===================
            if 'edges' in available_aspects:
                objects = self.streamAspect(uuid, 'edges', server, username, password)
                obj_items = (o for o in objects)
                for edge_item in obj_items:
                    add_this_edge = EdgesElement(json_obj=edge_item)

                    self.addEdge(add_this_edge)
                self.add_metadata_stub('edges')

            #===================
            # NODE ATTRIBUTES
            #===================
            if 'nodeAttributes' in available_aspects:
                objects = self.streamAspect(uuid, 'nodeAttributes', server, username, password)
                obj_items = (o for o in objects)
                for att in obj_items:
                    add_this_node_att = NodeAttributesElement(json_obj=att)

                    self.addNodeAttribute(add_this_node_att)
                self.add_metadata_stub('nodeAttributes')

            #===================
            # EDGE ATTRIBUTES
            #===================
            if 'edgeAttributes' in available_aspects:
                objects = self.streamAspect(uuid, 'edgeAttributes', server, username, password)
                obj_items = (o for o in objects)
                for att in obj_items:
                    add_this_edge_att = EdgeAttributesElement(json_obj=att)

                    self.addEdgeAttribute(add_this_edge_att)
                self.add_metadata_stub('edgeAttributes')

            #===================
            # CITATIONS
            #===================
            if 'citations' in available_aspects:
                objects = self.streamAspect(uuid, 'citations', server, username, password)
                obj_items = (o for o in objects)
                for cit in obj_items:
                    add_this_citation = CitationElement(json_obj=cit)

                    self.addCitation(add_this_citation)
                self.add_metadata_stub('citations')

            #===================
            # SUPPORTS
            #===================
            if 'supports' in available_aspects:
                objects = self.streamAspect(uuid, 'supports', server, username, password)
                obj_items = (o for o in objects)
                for sup in obj_items:
                    add_this_supports = SupportElement(json_obj=sup)

                    self.addSupport(add_this_supports)
                self.add_metadata_stub('supports')

            #===================
            # EDGE SUPPORTS
            #===================
            if 'edgeSupports' in available_aspects:
                objects = self.streamAspect(uuid, 'edgeSupports', server, username, password)
                obj_items = (o for o in objects)
                for add_this_edge_sup in obj_items:
                    self.addEdgeSupports(add_this_edge_sup)

                self.add_metadata_stub('edgeSupports')

            #===================
            # NODE CITATIONS
            #===================
            if 'nodeCitations' in available_aspects:
                objects = self.streamAspect(uuid, 'nodeCitations', server, username, password)
                obj_items = (o for o in objects)
                for node_cit in obj_items:
                    self.addNodeCitationsFromCX(node_cit)
                self.add_metadata_stub('nodeCitations')

            #===================
            # EDGE CITATIONS
            #===================
            if 'edgeCitations' in available_aspects:
                objects = self.streamAspect(uuid, 'edgeCitations', server, username, password)
                obj_items = (o for o in objects)
                for edge_cit in obj_items:
                    self.addEdgeCitationsFromCX(edge_cit)
                self.add_metadata_stub('edgeCitations')

            #===================
            # OPAQUE ASPECTS
            #===================
            for oa in opaque_aspects:
                objects = self.streamAspect(uuid, oa, server, username, password)
                obj_items = (o for o in objects)
                for oa_item in obj_items:
                    aspect_element = AspectElement(oa_item, oa)
                    self.addOpapqueAspect(aspect_element)
                    self.add_metadata_stub(oa)
        else:
            raise Exception('Server and uuid not specified')

    def create_from_cx(self, cx):
        if cx:
            #===================
            # METADATA
            #===================
            available_aspects = []
            for ae in (o for o in self.get_frag_from_list_by_key(cx, 'metaData')):
                available_aspects.append(ae.get(CX_CONSTANTS.METADATA_NAME))
                mde = MetaDataElement(json_obj=ae)
                self.addMetadata(mde)

            opaque_aspects = set(available_aspects).difference(known_aspects_min)

            #====================
            # NETWORK ATTRIBUTES
            #====================
            if 'networkAttributes' in available_aspects:
                objects = self.get_frag_from_list_by_key(cx, 'networkAttributes')
                obj_items = (o for o in objects)
                for network_item in obj_items:
                    add_this_network_attribute = NetworkAttributesElement(json_obj=network_item)

                    self.addNetworkAttribute(add_this_network_attribute)
                self.add_metadata_stub('networkAttributes')

            #===================
            # NODES
            #===================
            if 'nodes' in available_aspects:
                objects = self.get_frag_from_list_by_key(cx, 'nodes')
                obj_items = (o for o in objects)
                for node_item in obj_items:
                    add_this_node = NodesElement(json_obj=node_item)

                    self.addNode(add_this_node)
                self.add_metadata_stub('nodes')

            #===================
            # EDGES
            #===================
            if 'edges' in available_aspects:
                objects = self.get_frag_from_list_by_key(cx, 'edges')
                obj_items = (o for o in objects)
                for edge_item in obj_items:
                    add_this_edge = EdgesElement(json_obj=edge_item)

                    self.addEdge(add_this_edge)
                self.add_metadata_stub('edges')

            #===================
            # NODE ATTRIBUTES
            #===================
            if 'nodeAttributes' in available_aspects:
                objects = self.get_frag_from_list_by_key(cx, 'nodeAttributes')
                obj_items = (o for o in objects)
                for att in obj_items:
                    add_this_node_att = NodeAttributesElement(json_obj=att)

                    self.addNodeAttribute(add_this_node_att)
                self.add_metadata_stub('nodeAttributes')

            #===================
            # EDGE ATTRIBUTES
            #===================
            if 'edgeAttributes' in available_aspects:
                objects = self.get_frag_from_list_by_key(cx, 'edgeAttributes')
                obj_items = (o for o in objects)
                for att in obj_items:
                    add_this_edge_att = EdgeAttributesElement(json_obj=att)

                    self.addEdgeAttribute(add_this_edge_att)
                self.add_metadata_stub('edgeAttributes')

            #===================
            # CITATIONS
            #===================
            if 'citations' in available_aspects:
                objects = self.get_frag_from_list_by_key(cx, 'citations')
                obj_items = (o for o in objects)
                for cit in obj_items:
                    add_this_citation = CitationElement(json_obj=cit)

                    self.addCitation(add_this_citation)
                self.add_metadata_stub('citations')

            #===================
            # SUPPORTS
            #===================
            if 'supports' in available_aspects:
                objects = self.get_frag_from_list_by_key(cx, 'supports')
                obj_items = (o for o in objects)
                for sup in obj_items:
                    add_this_supports = SupportElement(json_obj=sup)

                    self.addSupport(add_this_supports)
                self.add_metadata_stub('supports')

            #===================
            # EDGE SUPPORTS
            #===================
            if 'edgeSupports' in available_aspects:
                objects = self.get_frag_from_list_by_key(cx, 'edgeSupports')
                obj_items = (o for o in objects)
                for add_this_edge_sup in obj_items:
                    self.addEdgeSupports(add_this_edge_sup)

                self.add_metadata_stub('edgeSupports')

            #===================
            # NODE CITATIONS
            #===================
            if 'nodeCitations' in available_aspects:
                objects = self.get_frag_from_list_by_key(cx, 'nodeCitations')
                obj_items = (o for o in objects)
                for node_cit in obj_items:
                    self.addNodeCitationsFromCX(node_cit)
                self.add_metadata_stub('nodeCitations')

            #===================
            # EDGE CITATIONS
            #===================
            if 'edgeCitations' in available_aspects:
                objects = self.get_frag_from_list_by_key(cx, 'edgeCitations')
                obj_items = (o for o in objects)
                for edge_cit in obj_items:
                    self.addEdgeCitationsFromCX(edge_cit)
                self.add_metadata_stub('edgeCitations')

            #===================
            # OPAQUE ASPECTS
            #===================
            for oa in opaque_aspects:
                objects = self.get_frag_from_list_by_key(cx, oa)
                obj_items = (o for o in objects)
                for oa_item in obj_items:
                    aspect_element = AspectElement(oa_item, oa)
                    self.addOpapqueAspect(aspect_element)
                    self.add_metadata_stub(oa)
        else:
            raise Exception('CX is empty')

    def get_frag_from_list_by_key(self, cx, key):
        for aspect in cx:
            if key in aspect:
                return aspect[key]

        return []

    def to_pandas(self):
        rows = []
        edge_items = None
        if sys.version_info.major == 3:
            edge_items = self.edges.items()
        else:
            edge_items = self.edges.iteritems()

        for k, v in edge_items:
            e_a = self.edgeAttributes.get(k)
            #==========================
            # PROCESS EDGE ATTRIBUTES
            #==========================
            add_this_dict = {}
            if e_a is not None:
                for e_a_item in e_a:
                    if type(e_a_item.getValues()) is list:
                        add_this_dict[e_a_item.getName()] = ','.join(str(e) for e in e_a_item.getValues())
                        add_this_dict[e_a_item.getName()] = '"' + add_this_dict[e_a_item.getName()] + '"'
                    else:
                        add_this_dict[e_a_item.getName()] = e_a_item.getValues()
            #================================
            # PROCESS SOURCE NODE ATTRIBUTES
            #================================
            s_a = self.nodeAttributes.get(v.getSource())
            if s_a is not None:
                for s_a_item in s_a:
                    if type(s_a_item.getValues()) is list:
                        add_this_dict['source_' + s_a_item.getName()] = ','.join(str(e) for e in s_a_item.getValues())
                        add_this_dict['source_' + s_a_item.getName()] = '"' + add_this_dict['source_' + s_a_item.getName()] + '"'
                    else:
                        add_this_dict['source_' + s_a_item.getName()] = s_a_item.getValues()

            #================================
            # PROCESS TARGET NODE ATTRIBUTES
            #================================
            t_a = self.nodeAttributes.get(v.getTarget())
            if t_a is not None:
                for t_a_item in t_a:
                    if type(t_a_item.getValues()) is list:
                        add_this_dict['target_' + t_a_item.getName()] = ','.join(str(e) for e in t_a_item.getValues())
                        add_this_dict['target_' + t_a_item.getName()] = '"' + add_this_dict['target_' + t_a_item.getName()] + '"'
                    else:
                        add_this_dict['target_' + t_a_item.getName()] = t_a_item.getValues()

            if add_this_dict:
                rows.append(dict(add_this_dict, source=self.nodes.get(v.getSource())._node_name, target=self.nodes.get(v.getTarget())._node_name, interaction=v._interaction))
            else:
                rows.append(dict(source=self.nodes.get(v.getSource())._node_name, target=self.nodes.get(v.getTarget())._node_name, interaction=v._interaction))

        nodeAttributeSourceTarget = []
        for n_a in self.nodeAttributeHeader:
            nodeAttributeSourceTarget.append('source_' + n_a)
            nodeAttributeSourceTarget.append('target_' + n_a)

        df_columns = ['source', 'interaction', 'target'] + list(self.edgeAttributeHeader) + nodeAttributeSourceTarget

        return_df = pd.DataFrame(rows, columns=df_columns)

        return return_df

    def add_metadata_stub(self, aspect_name):
        md = self.metadata.get(aspect_name)
        if md is None:
            mde = MetaDataElement(elementCount=0, properties=[], version='1.0', consistencyGroup=1, name=aspect_name)
            self.addMetadata(mde)

    def stream_cx(self):
        """Convert this network to a CX stream

        :return: The CX stream representation of this network.
        :rtype: io.BytesIO

        """
        cx = self.to_cx()

        if sys.version_info.major == 3:
            return io.BytesIO(json.dumps(cx).encode('utf-8'))
        else:
            return_bytes = None
            try:
                return_bytes = io.BytesIO(json.dumps(cx))
            except UnicodeDecodeError as err1:
                print "Detected invalid encoding. Trying latin-1 encoding."
                return_bytes = io.BytesIO(json.dumps(cx, encoding="latin-1"))
                print "Success"
            except Exception as err2:
                print err2.message

            return return_bytes

    def upload_to(self, server, username, password):
        """ Upload this network to the specified server to the account specified by username and password.

        :param server: The NDEx server to upload the network to.
        :type server: str
        :param username: The username of the account to store the network.
        :type username: str
        :param password: The password for the account.
        :type password: str
        :return: The UUID of the network on NDEx.
        :rtype: str

        Example:
            ndexGraph.upload_to('http://test.ndexbio.org', 'myusername', 'mypassword')
        """

        if server and 'http' not in server:
            server = 'http://' + server

        ndex = nc.Ndex2(server,username,password)
        save_this_cx = self.to_json()
        return ndex.save_new_network(save_this_cx)

    def upload_new_network_stream(self, server, username, password):
        response = ncs.postNiceCxStream(self)

        print response

    def update_to(self, uuid, server, username, password):
        """ Upload this network to the specified server to the account specified by username and password.

        :param server: The NDEx server to upload the network to.
        :type server: str
        :param username: The username of the account to store the network.
        :type username: str
        :param password: The password for the account.
        :type password: str
        :return: The UUID of the network on NDEx.
        :rtype: str

        Example:
            ndexGraph.upload_to('http://test.ndexbio.org', 'myusername', 'mypassword')
        """
        cx = self.to_json()
        ndex = nc.Ndex2(server,username,password)

        if(len(cx) > 0):
            if(cx[len(cx) - 1] is not None):
                if(cx[len(cx) - 1].get('status') is None):
                    # No STATUS element in the array.  Append a new status
                    cx.append({"status" : [ {"error" : "","success" : True} ]})
                else:
                    if(len(cx[len(cx) - 1].get('status')) < 1):
                        # STATUS element found, but the status was empty
                        cx[len(cx) - 1].get('status').append({"error" : "","success" : True})

            if sys.version_info.major == 3:
                stream = io.BytesIO(json.dumps(cx).encode('utf-8'))
            else:
                stream = io.BytesIO(json.dumps(cx))

            return ndex.update_cx_network(stream, uuid)
        else:
            raise IndexError("Cannot save empty CX.  Please provide a non-empty CX document.")

    def to_networkx(self):
        G = nx.Graph()

        if sys.version_info.major == 3:
            node_items = self.nodes.items()
            edge_items = self.edges.items()
        else:
            node_items = self.nodes.iteritems()
            edge_items = self.edges.iteritems()

        #============================
        # PROCESS NETWORK ATTRIBUTES
        #============================
        for net_a in self.networkAttributes:
            G.graph[net_a.getName()] = net_a.getValues()

        #================================
        # PROCESS NODE & NODE ATTRIBUTES
        #================================
        for k, v in node_items:
            node_attrs = {}
            n_a = self.nodeAttributes.get(k)
            if n_a:
                for na_item in n_a:
                    node_attrs[na_item.getName()] = na_item.getValues()
                    print v
                    my_name = v.getName()
                G.add_node(k, node_attrs, name=v.getName())

        #================================
        # PROCESS EDGE & EDGE ATTRIBUTES
        #================================
        for k, v in edge_items:
            e_a = self.edgeAttributes.get(k)
            add_this_dict = {}
            if e_a is not None:
                for e_a_item in e_a:
                    if type(e_a_item.getValues()) is list:
                        add_this_dict[e_a_item.getName()] = ','.join(str(e) for e in e_a_item.getValues())
                        add_this_dict[e_a_item.getName()] = '"' + add_this_dict[e_a_item.getName()] + '"'
                    else:
                        add_this_dict[e_a_item.getName()] = e_a_item.getValues()

                G.add_edge(v.getSource(), v.getTarget(), add_this_dict)
            else:
                G.add_edge(v.getSource(), v.getTarget())

        #================
        # PROCESS LAYOUT
        #================
        #if hasattr(networkx_G, 'pos'):
        #    G.pos = {node_dict[a] : b for a, b in networkx_G.pos.items()}

        return G

    def __str__(self):
        return json.dumps(self.to_json())

    def to_json(self):
        output_cx = [{"numberVerification": [{"longNumber": 281474976710655}]}]

        #=====================================================
        # IF THE @ID IS NOT NUMERIC WE NEED TO CONVERT IT TO
        # INT BY USING THE INDEX OF THE NON-NUMERIC VALUE
        #=====================================================
        if self.node_int_id_generator:
            self.node_id_lookup = list(self.node_int_id_generator)

        if self.metadata:
            #output_cx.append(self.generateAspect('metaData'))
            output_cx.append(self.generateMetadataAspect())
        if self.nodes:
            output_cx.append(self.generateAspect('nodes'))
        if self.edges:
            output_cx.append(self.generateAspect('edges'))
        if self.networkAttributes:
            output_cx.append(self.generateAspect('networkAttributes'))
        if self.nodeAttributes:
            output_cx.append(self.generateAspect('nodeAttributes'))
        if self.edgeAttributes:
            output_cx.append(self.generateAspect('edgeAttributes'))
        if self.citations:
            output_cx.append(self.generateAspect('citations'))
        if self.nodeCitations:
            output_cx.append(self.generateAspect('nodeCitations'))
        if self.edgeCitations:
            output_cx.append(self.generateAspect('edgeCitations'))
        if self.supports:
            output_cx.append(self.generateAspect('supports'))
        if self.edgeSupports:
            output_cx.append(self.generateAspect('edgeSupports'))
        if self.nodeSupports:
            output_cx.append(self.generateAspect('nodeSupports'))
        if self.opaqueAspects:
            for oa in self.opaqueAspects:
                output_cx.append({oa: self.opaqueAspects[oa]})
        if self.metadata:
            #===========================
            # UPDATE CONSISTENCY GROUP
            #===========================
            #self.updateConsistencyGroup()
            mt_a = self.generateMetadataAspect()

            #output_cx.append(self.generateAspect('metaData'))
            output_cx.append(self.generateMetadataAspect())

        #print json.dumps(output_cx)

        return output_cx

    def generateAspect(self, aspect_name):
        core_aspect = ['nodes', 'edges','networkAttributes', 'nodeAttributes', 'edgeAttributes', 'citations', 'metaData', 'supports']
        aspect_element_array = []
        element_count = 0
        element_id_max = 0

        use_this_aspect = None
        #=============================
        # PROCESS CORE ASPECTS FIRST
        #=============================
        if aspect_name in core_aspect:
            use_this_aspect = self.string_to_aspect_object(aspect_name)

        if use_this_aspect is not None:
            if type(use_this_aspect) is list:
                for item in use_this_aspect:
                    add_this_element = item.to_json()
                    id = add_this_element.get(CX_CONSTANTS.ID)
                    if id is not None and id > element_id_max:
                        element_id_max = id
                    aspect_element_array.append(add_this_element)
                    element_count +=1
            else:
                items = None
                if sys.version_info.major == 3:
                    items = use_this_aspect.items()
                else:
                    items = use_this_aspect.iteritems()

                for k, v in items:
                    if type(v) is list:
                        for v_item in v:
                            add_this_element = v_item.to_json()
                            id = add_this_element.get(CX_CONSTANTS.ID)
                            if id is not None and id > element_id_max:
                                element_id_max = id

                            if aspect_name == 'nodeAttributes':
                                if self.node_id_lookup:
                                    po_id = self.node_id_lookup.index(add_this_element.get(CX_CONSTANTS.PROPERTY_OF))
                                    add_this_element[CX_CONSTANTS.PROPERTY_OF] = po_id

                            aspect_element_array.append(add_this_element)
                            element_count +=1
                    else:
                        add_this_element = v.to_json()
                        id = add_this_element.get(CX_CONSTANTS.ID)
                        if aspect_name == 'nodes' and type(id) is str:
                            # CONVERT TO INT
                            id = self.node_id_lookup.index(id)
                            add_this_element[CX_CONSTANTS.ID] = id
                        if aspect_name == 'edges' and type(add_this_element.get(CX_CONSTANTS.EDGE_SOURCE_NODE_ID_OR_SUBNETWORK)) is str:
                            s_id = self.node_id_lookup.index(add_this_element.get(CX_CONSTANTS.EDGE_SOURCE_NODE_ID_OR_SUBNETWORK))
                            add_this_element[CX_CONSTANTS.EDGE_SOURCE_NODE_ID_OR_SUBNETWORK] = s_id

                        if aspect_name == 'edges' and type(add_this_element.get(CX_CONSTANTS.EDGE_TARGET_NODE_ID)) is str:
                            t_id = self.node_id_lookup.index(add_this_element.get(CX_CONSTANTS.EDGE_TARGET_NODE_ID))
                            add_this_element[CX_CONSTANTS.EDGE_TARGET_NODE_ID] = t_id

                        if id is not None and id > element_id_max:
                            element_id_max = id
                        aspect_element_array.append(add_this_element)
                        element_count +=1
        else:
            #===========================
            # PROCESS NON-CORE ASPECTS
            #===========================
            use_this_aspect = self.string_to_aspect_object(aspect_name)

            if use_this_aspect is not None:
                if type(use_this_aspect) is dict:
                    items = None
                    if sys.version_info.major == 3:
                        items = use_this_aspect.items()
                    else:
                        items = use_this_aspect.iteritems()

                    for k, v in items:
                        if aspect_name == 'edgeSupports':
                            if type(v) is list:
                                aspect_element_array.append({CX_CONSTANTS.PROPERTY_OF: [k], CX_CONSTANTS.SUPPORTS: v})
                            else:
                                aspect_element_array.append({CX_CONSTANTS.PROPERTY_OF: [k], CX_CONSTANTS.SUPPORTS: [v]})
                        else:
                            if type(v) is list:
                                aspect_element_array.append({CX_CONSTANTS.PROPERTY_OF: [k], CX_CONSTANTS.CITATIONS: v})
                            else:
                                aspect_element_array.append({CX_CONSTANTS.PROPERTY_OF: [k], CX_CONSTANTS.CITATIONS: [v]})
                        element_count +=1
                else:
                    raise Exception('Citation was not in json format')
            else:
                return None

        aspect = {aspect_name: aspect_element_array}
        #============================================
        # UPDATE METADATA ELEMENT COUNTS/ID COUNTER
        #============================================
        md = self.metadata.get(aspect_name)
        if md is not None:
            md.setElementCount(element_count)
            md.setIdCounter(element_id_max)
            #md.incrementConsistencyGroup()
        elif aspect_name != 'metaData':
            mde = MetaDataElement(elementCount=element_count, properties=[], version='1.0', consistencyGroup=1, name=aspect_name)

            if element_id_max != 0:
                mde.setIdCounter(element_id_max)

            self.addMetadata(mde)

        #print('%s ELEMENT COUNT: %s, MAX ID: %s' % (aspect_name, str(element_count), str(element_id_max)))
        return aspect

    def generateMetadataAspect(self):
        aspect_element_array = []
        element_count = 0
        element_id_max = 0

        use_this_aspect = self.string_to_aspect_object('metaData')

        if use_this_aspect is not None:
            if sys.version_info.major == 3:
                items = use_this_aspect.items()
            else:
                items = use_this_aspect.iteritems()

            for k, v in items:
                add_this_element = v.to_json()
                id = add_this_element.get(CX_CONSTANTS.ID)

                if id is not None and id > element_id_max:
                    element_id_max = id
                aspect_element_array.append(add_this_element)
                element_count +=1

        aspect = {'metaData': aspect_element_array}

        return aspect

    def handleMetadataUpdate(self, aspect_name):
        aspect = self.string_to_aspect_object(aspect_name)

        #return_metadata = {
        #    CX_CONSTANTS.CONSISTENCY_GROUP: consistency_group,
        #    CX_CONSTANTS.ELEMENT_COUNT: 1,
        #    CX_CONSTANTS.METADATA_NAME: "@context",
        #    CX_CONSTANTS.PROPERTIES: [],
        #    CX_CONSTANTS.VERSION: "1.0"
        #}

    def updateConsistencyGroup(self):
        consistency_group = 1
        if self.metadata:
            for mi_k, mi_v in self.metadata.items():
                cg = mi_v.getConsistencyGroup()
                if cg > consistency_group:
                    consistency_group = cg

            consistency_group += 1 # bump the consistency group up by one

            for mi_k, mi_v in self.metadata.items():
                #print mi_k
                #print mi_v
                mi_v.setConsistencyGroup(consistency_group)

    def generate_metadata(self, G, unclassified_cx):
        #if self.metadata:
        #    for k, v in self.metadata.iteritems():


        return_metadata = []
        consistency_group = 1
        if(self.metadata_original is not None):
            for mi in self.metadata_original:
                if(mi.get("consistencyGroup") is not None):
                    if(mi.get("consistencyGroup") > consistency_group):
                        consistency_group = mi.get("consistencyGroup")
                else:
                    mi['consistencyGroup'] = 0

            consistency_group += 1 # bump the consistency group up by one

            #print("consistency group max: " + str(consistency_group))

        # ========================
        # @context metadata
        # ========================
        if self.namespaces:
            return_metadata.append(
                {
                    "consistencyGroup": consistency_group,
                    "elementCount": 1,
                    "name": "@context",
                    "properties": [],
                    "version": "1.0"
                }
            )

        #========================
        # Nodes metadata
        #========================
        node_ids = [n[0] for n in G.nodes_iter(data=True)]
        if(len(node_ids) < 1):
            node_ids = [0]
        return_metadata.append(
            {
                "consistencyGroup" : consistency_group,
                "elementCount" : len(node_ids),
                "idCounter": max(node_ids),
                "name" : "nodes",
                "properties" : [ ],
                "version" : "1.0"
            }
        )

        #========================
        # Edges metadata
        #========================
        edge_ids = [e[2]for e in G.edges_iter(data=True, keys=True)]
        if(len(edge_ids) < 1):
            edge_ids = [0]
        return_metadata.append(
            {
                "consistencyGroup" : consistency_group,
                "elementCount" : len(edge_ids),
                "idCounter": max(edge_ids),
                "name" : "edges",
                "properties" : [ ],
                "version" : "1.0"
            }
        )

        #=============================
        # Network Attributes metadata
        #=============================
        if(len(G.graph) > 0):
            return_metadata.append(
                {
                    "consistencyGroup" : consistency_group,
                    "elementCount" : len(G.graph),
                    "name" : "networkAttributes",
                    "properties" : [ ],
                    "version" : "1.0"
                }
            )

        #===========================
        # Node Attributes metadata
        #===========================
        #id_max = 0
        attr_count = 0
        for node_id , attributes in G.nodes_iter(data=True):
            for attribute_name in attributes:
                if attribute_name != "name" and attribute_name != "represents":
                    attr_count += 1



        #
        # for n, nattr in G.nodes(data=True):
        #     if(bool(nattr)):
        #         attr_count += len(nattr.keys())
        #
        #     if(n > id_max):
        #         id_max = n

        if(attr_count > 0):
            return_metadata.append(
                {
                    "consistencyGroup" : consistency_group,
                    "elementCount" : attr_count,
                    #"idCounter": id_max,
                    "name" : "nodeAttributes",
                    "properties" : [ ],
                    "version" : "1.0"
                }
            )

        #===========================
        # Edge Attributes metadata
        #===========================
        #id_max = 0
        attr_count = 0
        for s, t, id, a in G.edges(data=True, keys=True):
            if(bool(a)):
                for attribute_name in a:
                    if attribute_name != "interaction":
                        attr_count += 1
                #attr_count += len(a.keys())

            # if(id > id_max):
            #     id_max = id

        if(attr_count > 0):
            return_metadata.append(
                {
                    "consistencyGroup" : consistency_group,
                    "elementCount" : attr_count,
                    #"idCounter": id_max,
                    "name" : "edgeAttributes",
                    "properties" : [ ],
                    "version" : "1.0"
                }
            )

        #===========================
        # cyViews metadata
        #===========================
        if self.view_id != None:
            return_metadata.append(
                {
                    "elementCount": 1,
                    "name": "cyViews",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )

        #===========================
        # subNetworks metadata
        #===========================
        if self.subnetwork_id != None:
            return_metadata.append(
                {
                    "elementCount": 1,
                    "name": "subNetworks",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )

        #===========================
        # networkRelations metadata
        #===========================
        if self.subnetwork_id != None and self.view_id != None:
            return_metadata.append(
                {
                    "elementCount": 2,
                    "name": "networkRelations",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )

        #===========================
        # citations and supports metadata
        #===========================
        if len(self.support_map) > 0:
            return_metadata.append(
                {
                    "elementCount": len(self.support_map),
                    "name": "supports",
                    "properties": [],
                    "idCounter": max(self.support_map.keys()),
                    "consistencyGroup" : consistency_group
                }
            )

        if len(self.node_support_map) > 0:
            return_metadata.append(
                {
                    "elementCount": len(self.node_support_map),
                    "name": "nodeSupports",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )
        if len(self.edge_support_map) > 0:
            return_metadata.append(
                {
                    "elementCount": len(self.edge_support_map),
                    "name": "edgeSupports",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )

        if len(self.citation_map) > 0:
            return_metadata.append(
                {
                    "elementCount": len(self.citation_map),
                    "name": "citations",
                    "properties": [],
                    "idCounter": max(self.citation_map.keys()),
                    "consistencyGroup" : consistency_group
                }
            )

        if len(self.node_citation_map) > 0:
            return_metadata.append(
                {
                    "elementCount": len(self.node_citation_map),
                    "name": "nodeCitations",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )

        if len(self.edge_citation_map) > 0:
            return_metadata.append(
                {
                    "elementCount": len(self.edge_citation_map),
                    "name": "edgeCitations",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )

        if len(self.function_term_map) > 0:
            return_metadata.append(
                {
                    "elementCount": len(self.function_term_map),
                    "name": "functionTerms",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )

        if len(self.reified_edges) > 0:
            return_metadata.append(
                {
                    "elementCount": len(self.reified_edges),
                    "name": "reifiedEdges",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )

        #===========================
        # ndexStatus metadata
        #===========================
        return_metadata.append(
            {
                "consistencyGroup": consistency_group,
                "elementCount": 1,
                "name": "ndexStatus",
                "properties": [],
                "version": "1.0"
            }
        )

        #===========================
        # cartesianLayout metadata
        #===========================
        if self.pos and len(self.pos) > 0:
            return_metadata.append(
                {
                    "consistencyGroup": consistency_group,
                    "elementCount": len(self.pos),
                    "name": "cartesianLayout",
                    "properties": [],
                    "version": "1.0"
                }
            )

        #===========================
        # OTHER metadata
        #===========================
        for asp in self.unclassified_cx:
            try:
                aspect_type = asp.iterkeys().next()
                if(aspect_type == "visualProperties"
                   or aspect_type == "cyVisualProperties"
                   or aspect_type == "@context"):
                    return_metadata.append(
                        {
                            "consistencyGroup" : consistency_group,
                            "elementCount":len(asp[aspect_type]),
                            "name":aspect_type,
                            "properties":[]
                         }
                    )
            except Exception as e:
                print(e.message)


        #print {'metaData': return_metadata}

        return [{'metaData': return_metadata}]

    def string_to_aspect_object(self, aspect_name):
        if aspect_name == 'metaData':
            return self.metadata
        elif aspect_name == 'nodes':
            return self.nodes
        elif aspect_name == 'edges':
            return self.edges
        elif aspect_name == 'networkAttributes':
            return self.networkAttributes
        elif aspect_name == 'nodeAttributes':
            return self.nodeAttributes
        elif aspect_name == 'edgeAttributes':
            return self.edgeAttributes
        elif aspect_name == 'citations':
            return self.citations
        elif aspect_name == 'nodeCitations':
            return self.nodeCitations
        elif aspect_name == 'edgeCitations':
            return self.edgeCitations
        elif aspect_name == 'edgeSupports':
            return self.edgeSupports
        elif aspect_name == 'supports':
            return self.supports

    def streamAspect(self, uuid, aspect_name, server, username, password):
        if 'http' not in server:
            server = 'http://' + server
        if aspect_name == 'metaData':
            print(server + '/v2/network/' + uuid + '/aspect')

            s = requests.session()
            if username and password:
                # add credentials to the session, if available
                s.auth = (username, password)
            md_response = s.get(server + '/v2/network/' + uuid + '/aspect')
            json_response = md_response.json()
            return json_response.get('metaData')
        else:
            if username and password:
                base64string = base64.b64encode('%s:%s' % (username, password))
                request = Request(server + '/v2/network/' + uuid + '/aspect/' + aspect_name, headers={"Authorization": "Basic " + base64.encodestring(username + ':' + password).replace('\n', '')})
            else:
                request = Request(server + '/v2/network/' + uuid + '/aspect/' + aspect_name)
            try:
                urlopen_result = urlopen(request) #'http://dev2.ndexbio.org/v2/network/' + uuid + '/aspect/' + aspect_name)
            except HTTPError as e:
                print(e.code)
                return []
            except URLError as e:
                print('Other error')
                print('URL Error %s' % e.message())
                return []

            return ijson.items(urlopen_result, 'item')

class DecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return float(o)
        return super(DecimalEncoder, self).default(o)



__author__ = 'aarongary'

from . import CX_CONSTANTS

class CXSimpleAttribute():
    def __init__(self, p=None):
        self.ID = CX_CONSTANTS.get('ID')
        self.INTERACTION = 'i'
        self.name = None
        self.value = None
        self.dataType = None

        if p is not None:
            self.name = p.getPredicateString()
            self.value = p.getValue()
            self.dataType = p.getDataType()

    def getName(self):
        return self.name

	def setName(self,name):
		self.name = name

	def getValue(self):
		return self.value

	def setValue(self, value):
		self.value = value

	def getDataType(self):
		return self.dataType

	def setDataType(self, dataType):
		self.dataType = dataType


'''
	@JsonProperty("n")
	private String name;

	@JsonProperty("v")
	private String value;

	@JsonProperty("t")
	private String dataType;

	public CXSimpleAttribute() {
	}

	public CXSimpleAttribute(NdexPropertyValuePair p ) {
		name = p.getPredicateString();
		value = p.getValue();
		dataType = p.getDataType();
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public String getValue() {
		return value;
	}

	public void setValue(String value) {
		this.value = value;
	}

	public String getDataType() {
		return dataType;
	}

	public void setDataType(String dataType) {
		this.dataType = dataType

'''


import ndex2.client as nc
from ndex2.create_aspect import domain
import json
import time

ndex_host = "http://dev2.ndexbio.org"
username_1 = "drh"
password_1 = "drh"

def get_network(ndex, network_id):
    response = ndex.get_network_as_cx_stream(example_network_1_id)
    return response.json()

def network_id_from_uri(network_uri):
    print("network URI: %s" % network_uri)
    return network_uri.rpartition('/')[2]

def update_network_properties(ndex, network_summary, property_dict):
    # This function handles two cases:
    # - there are no subnetworks, no properties have a specified subnetwork id
    # - there is one subnetwork, all properties have that subnetwork id
    # Otherwise, raise an exception
    #
    # new properties are added
    # existing properties with the same names as in the property_dict are updated.
    # other existing properties are preserved
    #
    # Cases to be handled by explicit calls to set_network_properties, not in this function:
    # ** adding / removing items from list values
    # ** property removal
    #
    # python data types in the property_dict are mapped to CX PropertyValuePair objects
    # using ndex.create_aspect.domain to determine the data type
    subnetwork_id = None
    property_list = []
    network_id = network_summary["externalId"]
    if "properties" in network_summary:
        property_list = network_summary["properties"]
    if "subnetworkIds" in network_summary:
        if len(network_summary["subnetworkIds"]) > 1:
            raise Exception("property setting for networks with multiple subnetworks is not handled, use the lower-level set_network_properties directly")
        subnetwork_id = network_summary["subnetworkIds"][0]
        for property in property_list:
            sid = property.get("subNetworkId")
            if not sid:
                raise Exception("property setting for networks with mixed subnetwork and base network properties is not handled")
            if sid != subnetwork_id:
                raise Exception("existing property on subnetwork %s but subnetwork %s is the only one specified in the network summary" % (sid, subnetworkId))

    for name, value in property_dict.iteritems():
        update_cx_property_list(name, value, subnetwork_id, property_list)

    ndex.set_network_properties(network_id, property_list)

def update_cx_property_list(name, value, subnetwork_id, property_list):
    data_type = domain(value)
    if data_type == "unknown":
        raise Exception("in attempt to update property %s with value %s, no CX data type could be determined" % (name, value))
    for property in property_list:
        if property["predicateString"] == name:
            property["value"] = str(value)
            property["dataType"] = data_type
            return
    property_list.append({
        "predicateString": name,
        "subNetworkId": subnetwork_id,
        "value": str(value),
        "dataType": data_type
    })

example_network_1_id = "b612d677-c714-11e6-b48c-0660b7976219"

ndex = nc.Ndex2(host=ndex_host, username=username_1, password=password_1,debug=True)

example_network_1_summary = ndex.get_network_summary(example_network_1_id)

print(json.dumps(example_network_1_summary, indent=4))

# Get the network

example_network_1 = get_network(ndex, example_network_1_id)

# Save it as a new network

test_network_1_uri = ndex.save_new_network(example_network_1)

test_network_1_id = network_id_from_uri(test_network_1_uri)

test_network_1_summary = ndex.get_network_summary(test_network_1_id)

print(json.dumps(test_network_1_summary, indent=4))

#----------------------------
# Update visibility

time.sleep(3)
ndex.make_network_public(test_network_1_id)
print("network is public")


#----------------------------
#  Update Properties

test_property_dict = {"string_test": "bar", "int_test": 23, "float_test": 1.001}

update_network_properties(ndex, test_network_1_summary, test_property_dict)

#ndex.set_network_properties(test_network_1_id, test_properties)


# Get the summary
time.sleep(5)
test_network_1_summary = ndex.get_network_summary(test_network_1_id)

print(json.dumps(test_network_1_summary, indent=4))

#----------------------------
# Update the profile

test_profile = {"version": "3.69",
               "name": "test - The Heart Machine",
               "description": "<h1>TEST</h1><div>The Heart Machine reads new publications on the molecular mechanisms governing cardiac cells automatically and assembles a model from them using INDRA.<br/></div>"}

ndex.update_network_profile(test_network_1_id, test_profile)

# Get the summary

test_network_1_summary = ndex.get_network_summary(test_network_1_id)

print(json.dumps(test_network_1_summary, indent=4))



# Get the

# Clean up


__author__ = 'aarongary'

from . import CX_CONSTANTS

class AspectElement(object):
    def __init__(self, element, aspect_name):
        self.serialVersionUID = 1
        self.ASPECT_NAME = aspect_name
        self.aspect_element = element

    def getAspectName(self):
        return self.ASPECT_NAME

    def setAspectName(self, name):
        self.ASPECT_NAME = name

    def getAspectElement(self):
        return self.aspect_element

    def setAspectElement(self, ae):
        self.aspect_element = ae



__author__ = 'aarongary'

import unittest
import ndex2
from ndex2.networkn import  NdexGraph
import json

class NetworkNTests(unittest.TestCase):
    #==============================
    # TEST LARGE NETWORK
    #==============================
    def test_data_to_type(self):
        self.assertTrue(self, ndex2.networkn.data_to_type('true','boolean'))
        print(type(ndex2.networkn.data_to_type('1.3','double')))
        print(type(ndex2.networkn.data_to_type('1.3','float')))
        print(type(ndex2.networkn.data_to_type('1','integer')))
        print(type(ndex2.networkn.data_to_type('1','long')))
        print(type(ndex2.networkn.data_to_type('1','short')))
        print(type(ndex2.networkn.data_to_type('1','string')))
        list_of_boolean = type(ndex2.networkn.data_to_type('["true","false"]','list_of_boolean'))
        print(list_of_boolean)

        list_of_double = ndex2.networkn.data_to_type('[1.3,1.4]','list_of_double')
        print(list_of_double)

        list_of_float = ndex2.networkn.data_to_type('[1.3,1.4]','list_of_float')
        print(list_of_float)

        list_of_integer = ndex2.networkn.data_to_type('[1,4]','list_of_integer')
        print(list_of_integer)

        list_of_long = ndex2.networkn.data_to_type('[1,4]','list_of_long')
        print(list_of_long)

        list_of_short = ndex2.networkn.data_to_type('[1,4]','list_of_short')
        print(list_of_short)

        list_of_string = ndex2.networkn.data_to_type(['abc'],'list_of_string')
        print(list_of_string)

    def test_edge_float_type(self):
        G = NdexGraph()
        n_a = G.add_new_node(name='A')
        n_b = G.add_new_node(name='B')
        n_c = G.add_new_node(name='C')
        n_d = G.add_new_node(name='D')

        e_1 = G.add_edge(n_a, n_b, 10, {'weight': 1.234})
        e_2 = G.add_edge(n_a, n_c, 11, {'weight': 2.554})
        e_3 = G.add_edge(n_a, n_d, 12, {'weight': 5.789})
        e_4 = G.add_edge(n_b, n_c, 13, {'weight': 2.011})
        e_5 = G.add_edge(n_b, n_d, 14, {'weight': 7.788})

        print json.dumps(G.to_cx())
        G.upload_to('http://dev.ndexbio.org', 'scratch', 'scratch')


if __name__ == '__main__':
    unittest.main()


__author__ = 'aarongary'

import json
from nicecxModel.cx import CX_CONSTANTS

class NodesElement(object):
    def __init__(self, id=None, node_name=None, node_represents=None, json_obj=None):
        self.ASPECT_NAME = 'nodes'

        self._node_name = node_name
        self._node_represents = node_represents

        if json_obj is not None:
            if type(json_obj) is dict:
                self._node_name = json_obj.get(CX_CONSTANTS.NAME)
                self._node_represents = json_obj.get(CX_CONSTANTS.NODE_REPRESENTS)
                if json_obj.get(CX_CONSTANTS.ID) is not None:
                    self._id = json_obj.get(CX_CONSTANTS.ID)
            else:
                raise Exception('NodesElement json input provided was not of type json object.')

        if id is None:
            self._id = node_name
        else:
            self._id = id


    def getAspectName(self):
        return self.ASPECT_NAME

    def getId(self):
        return self._id

    def setId(self, id):
        self._id = id

    def getNodeRepresents(self):
        return self._node_represents

    def setNodeRepresents(self, represents):
        self._node_represents = represents

    def getName(self):
        return self._node_name

    def setNodeName(self, node_name ):
        self._node_name = node_name

    def __eq__ (self, other):
        return self == other or self._id == other._id

    def __str__(self):
        return json.dumps(self.to_json())

    def to_json(self):
        if self._id == -1:
            raise Exception('Edge element does not have a valid ID.  Unable to process this edge - ' + self._node_name)

        node_dict = {
            CX_CONSTANTS.ID: self._id,
            CX_CONSTANTS.NAME: self._node_name
        }

        if self._node_represents is not None:
            node_dict[CX_CONSTANTS.NODE_REPRESENTS] = self._node_represents

        return node_dict






import networkx as nx
from networkx.classes.multidigraph import MultiDiGraph
import ndex2.create_aspect as ca
import io
import json
import copy
import ndex2.client as nc
from time import time
from six import string_types
import sys
import warnings
from ndex2 import deprecation_message

#NDEXGRAPH_RESERVED_ATTRIBUTES = [
#    "subnetwork_id"
#    "view_id",
#    "max_node_id",
#    "max_edge_id",
#    "max_citation_id",
#    "max_support_id",
#    "pos",
#    "unclassified_cx",
#    "metadata_original",
#    "status",
#    "citation_map",
#    "node_citation_map",
#    "edge_citation_map",
#    "support_map",
#    "node_support_map",
#    "edge_support_map",
#    "self.edgemap"
#]

# TODO - Deprecate all methods. Direct users to nice cx
def parse_attribute(attribute):
    value = attribute['v']
    if 'd' in attribute:
        d = attribute['d']
        value = data_to_type(value,d)
    return value


def data_to_type(data, data_type):
    return_data = None

    if(type(data) is str):
        data = data.replace('[', '').replace(']','')
        if('list_of' in data_type):
            data = data.split(',')

    if data_type == "boolean":
        if(type(data) is str):
            return_data = data.lower() == 'true'
        else:
            return_data = bool(data)
    elif data_type == "byte":
        return_data = str(data).encode()
    elif data_type == "char":
        return_data = str(data)
    elif data_type == "double":
        return_data = float(data)
    elif data_type == "float":
        return_data = float(data)
    elif data_type == "integer":
        return_data = int(data)
    elif data_type == "long":
        return_data = int(data)
    elif data_type == "short":
        return_data = int(data)
    elif data_type == "string":
        return_data = str(data)
    elif data_type == "list_of_boolean":
        # Assumption: if the first element is a string then so are the rest...
        if(type(data[0]) is str):
            return_data = [s.lower() == 'true' for s in data]
        else:
            return_data = [bool(s) for s in data]
    elif data_type == "list_of_byte":
        return_data = [bytes(s) for s in data]
    elif data_type == "list_of_char":
        return_data = [str(s) for s in data]
    elif data_type == "list_of_double":
        return_data = [float(s) for s in data]
    elif data_type == "list_of_float":
        return_data = [float(s) for s in data]
    elif data_type == "list_of_integer":
        return_data = [int(s) for s in data]
    elif data_type == "list_of_long":
        return_data = [int(s) for s in data]
    elif data_type == "list_of_short":
        return_data = [int(s) for s in data]
    elif data_type == "list_of_string":
        return_data = [str(s) for s in data]
    else:
        return None

    return return_data

class NdexGraph (MultiDiGraph):
    """A graph compatible with NDEx"""

    def __init__(self, cx=None, server=None, username=None, password=None, uuid=None, networkx_G=None, data=None, **attr):
        """There are generally four ways to create a graph.

            1. An empty graph. G = NdexGraph()
            2. Using a cx dictionary. G = NdexGraph(cx)
            3. Loading it from an NDEx server.
                G = NdexGraph(server='http://test.ndexbio.org' uuid='983a2b93-2c55-11e6-a7c5-0630eb0972a1')
            4. Just like any other NetworkX MultiDiGraph().
            5. From an existing networkx graph. G = NdexGraph(networkx_G=networkx_G)

        """
        warnings.warn(deprecation_message)
        MultiDiGraph.__init__(self, data, **attr)
        self.subnetwork_id = None
        self.view_id = None
        self.max_node_id = None
        self.max_edge_id = None
        self.max_citation_id = None
        self.max_support_id = None
        self.pos = {}
        self.unclassified_cx = []
        self.metadata_original = None
        self.status = {'status': [{'error': '','success': True}]} #Added because status is now required
        self.citation_map = {}
        self.node_citation_map = {}
        self.citation_reference_map = {}
        self.edge_citation_map = {}
        self.support_map = {}
        self.support_reference_map = {}
        self.node_support_map = {}
        self.edge_support_map = {}
        self.function_term_map = {}
        self.reified_edges = {}
        self.provenance = None
        self.namespaces = {}
     #   self.edge_type_map = {}  # stores the mapping from cx edgeId to its 'i' attribute
     #   self.node_prepresent_map ={} #stores the mapping from cx nodeId to its represents

        # Maps edge ids to node ids. e.g. { edge1: (source_node, target_node), edge2: (source_node, target_node) }
        self.edgemap = {}

        if networkx_G is not None:
            node_id_x = 0
            node_dict_x = {}
            self.max_edge_id = 0
            for node_name, node_attr in networkx_G.nodes_iter(data=True):
                if node_attr.has_key('name'):
                    self.add_node(node_id_x, node_attr)
                else:
                    self.add_node(node_id_x, node_attr, name=node_name)
                node_dict_x[node_name] = node_id_x
                node_id_x += 1
            for s, t, edge_attr in networkx_G.edges_iter(data=True):
                self.add_edge(node_dict_x[s], node_dict_x[t], self.max_edge_id, edge_attr)
                self.max_edge_id += 1

            if hasattr(networkx_G, 'pos'):
                self.pos = {node_dict_x[a] : b for a, b in networkx_G.pos.items()}
                self.subnetwork_id = 1
                self.view_id = 1

            if isinstance(networkx_G, nx.MultiGraph):
                for s, t, key, data in networkx_G.edges_iter(keys=True, data=True):
                    self.add_edge(s, t, key, data)
            else:
                for s, t, data in networkx_G.edges_iter(data=True):
                    self.add_edge(s, t, self.max_edge_id, data)
                    self.max_edge_id += 1
            return

        if not cx and server and uuid:
            ndex = nc.Ndex2(server,username,password)
            cx = ndex.get_network_as_cx_stream(uuid).json()
            if not cx:
                raise RuntimeError("Failed to retrieve network with uuid " + uuid + " from " + server)
            else:
                metadata_temp = (item for item in cx if item.get("metaData") is not None).next()
                if(metadata_temp is not None):
                    self.metadata_original = metadata_temp["metaData"]

        # If there is no CX to process, just return.
        if cx == None:
            return

        # First pass, get information about subnetworks.
        for aspect in cx:
            if 'status' in aspect :
                if aspect['status'][0]['success']:
                    continue
                else:
                    raise RuntimeError("Error in CX status aspect: " + aspect['status'][0]['error'])
            if "numberVerification" in aspect:
                # new status and numberVerification will be added when the network is output to_cx
                continue
            if 'subNetworks' in aspect:
                for subnetwork in aspect.get('subNetworks'):
                    id = subnetwork.get('@id')
                    if self.subnetwork_id != None:
                        raise ValueError("networkn does not support collections!")
                    self.subnetwork_id = id
            elif 'cyViews' in aspect:
                for cyViews in aspect.get('cyViews'):
                    id = cyViews.get('@id')
                    if self.view_id != None:
                        raise ValueError("networkn does not support more than one view!")
                    self.view_id = id
            elif 'metaData' in aspect:
                self.metadata_original = aspect["metaData"]
                # Strip metaData
                continue
            elif 'provenanceHistory' in aspect:
                elements = aspect['provenanceHistory']
                if len(elements) > 0:
                    if len(elements)>1 or self.provenance :
                        raise RuntimeError('profenanceHistory aspect can only have one element.')
                    else :
                        self.provenance = elements[0]
            elif '@context' in aspect :
                elements = aspect['@context']
                if len(elements) > 0:
                    if  len(elements) > 1 or self.namespaces:
                        raise RuntimeError('@context aspect can only have one element')
                    else :
                        self.namespaces = elements[0]
            # TODO elif if it's an aspect we want to keep out we put an elif for that aspect
            else:
                self.unclassified_cx.append(aspect)

            cx = self.unclassified_cx

        # Second pass, just build basic graph.
        self.unclassified_cx = []
        for aspect in cx:
            if 'nodes' in aspect:
                for node in aspect.get('nodes'):
                    id = node.get('@id')
                    name = node['n'] if 'n' in node else None
                    if name:
                        self.add_node(id, name=name)
                    else:
                        self.add_node(id)
                    represents = node.get('r') if 'r' in node else None
                    if represents:
                        self.node.get(id)['represents'] = represents

            elif 'edges' in aspect:
                for edge in aspect.get('edges'):
                    id = edge.get('@id')
                    interaction = edge['i'] if 'i' in edge else None
                    s = edge['s']
                    t = edge['t']
                    self.edgemap[id] = (s, t)
                    if interaction:
                        self.add_edge(s, t, key=id, interaction=interaction)
                    else:
                        self.add_edge(s, t, key=id)
            else:
                self.unclassified_cx.append(aspect)
        cx = self.unclassified_cx

        # Third pass, handle attributes
        # Notes. Not handled, datatypes.
        self.unclassified_cx = []
        for aspect in cx:
            if 'networkAttributes' in aspect:
                for networkAttribute in aspect['networkAttributes']:
                    name = networkAttribute['n']
                    # special: ignore selected
           #         if name == 'selected':
           #             continue
                    value = parse_attribute(networkAttribute)
                    value = networkAttribute['v']
                    if value is not None:
                        if 's' in networkAttribute or name not in self.graph:
                            self.graph[name] = value

            elif 'nodeAttributes' in aspect:
                for nodeAttribute in aspect['nodeAttributes']:
                    id = nodeAttribute['po']
                    name = nodeAttribute['n']
                    value = parse_attribute(nodeAttribute)
                    if value is not None:
                        if 's' in nodeAttribute or name not in self.node[id]:
                            self.node[id][name] = value

            elif 'edgeAttributes' in aspect:
                for edgeAttribute in aspect['edgeAttributes']:
                    id = edgeAttribute['po']
                    s, t = self.edgemap[id]
                    name = edgeAttribute['n']
                    value = parse_attribute(edgeAttribute)
                    if value is not None:
                        if 's' in edgeAttribute or name not in self[s][t][id]:
                            self[s][t][id][name] = value
            else:
                self.unclassified_cx.append(aspect)

        cx = self.unclassified_cx
        self.unclassified_cx = []

        # Fourth pass, node locations
        self.pos = {}
        for aspect in cx:
            if 'cartesianLayout' in aspect:
                for nodeLayout in aspect['cartesianLayout']:
                    id = nodeLayout['node']
                    x = nodeLayout['x']
                    y = nodeLayout['y']
                    self.pos[id] = [x,y]
            else:
                self.unclassified_cx.append(aspect)

        cx = self.unclassified_cx
        self.unclassified_cx = []

        # Fifth pass, citations
        for aspect in cx:
            if 'citations' in aspect:
                for citation in aspect['citations']:
                    attributes = copy.deepcopy(citation)
                    attributes.pop("@id")
                    self.citation_map[citation["@id"]] = attributes
                    self.citation_reference_map[citation["@id"]] = 0
            else:
                self.unclassified_cx.append(aspect)

        cx = self.unclassified_cx
        self.unclassified_cx = []

        # Sixth pass, supports, the attributes include the mapping to their citation (if any) by its id
        for aspect in cx:
            if 'supports' in aspect:
                for support in aspect['supports']:
                    attributes = copy.deepcopy(support)
                    attributes.pop("@id")
                    self.support_map[support["@id"]] = attributes
                    self.support_reference_map[support["@id"]] = 0

            else:
                self.unclassified_cx.append(aspect)

        cx = self.unclassified_cx
        self.unclassified_cx = []
        # Seventh pass, map supports and citations to nodes and edges
        for aspect in cx:
            if 'nodeCitations' in aspect:
                for node_citation in aspect['nodeCitations']:
                    for node in node_citation["po"]:
                        self.node_citation_map[node] = node_citation["citations"]
                        for citation_id in node_citation["citations"]:
                            cit_ref = self.citation_reference_map.get(citation_id)
                            if(cit_ref is not None):
                                self.citation_reference_map[citation_id] += 1
            elif 'edgeCitations' in aspect:
                for edge_citation in aspect['edgeCitations']:
                    for edge in edge_citation["po"]:
                        edge_citation_added = False
                        for citation_id in edge_citation["citations"]:
                            citation_tmp = self.citation_map.get(citation_id)
                            if citation_tmp is not None and isinstance(citation_tmp, dict):
                                self.add_citation_to_edge(edge, citation_tmp.get('dc:identifier'))
                                edge_citation_added = True
                            if citation_tmp is not None and isinstance(citation_tmp, basestring):
                                self.add_citation_to_edge(edge, citation_tmp)
                                edge_citation_added = True

                            cit_ref = self.citation_reference_map.get(citation_id)
                            if(cit_ref is not None):
                                self.citation_reference_map[citation_id] += 1

                        if not edge_citation_added:
                            self.edge_citation_map[edge] = edge_citation["citations"]

            elif 'nodeSupports' in aspect:
                for node_support in aspect['nodeSupports']:
                    for node_sup_po in node_support["po"]:
                        self.node_support_map[node_sup_po] = node_support["supports"]
                        for supports_id in node_support["supports"]:
                            sup_ref = self.support_reference_map.get(supports_id)
                            if(sup_ref is not None):
                                self.support_reference_map[supports_id] += 1
            elif 'edgeSupports' in aspect:
                for edge_support in aspect['edgeSupports']:
                    for edge_sup in edge_support["po"]:
                        self.edge_support_map[edge_sup] = edge_support["supports"]
                        for supports_id in edge_support["supports"]:
                            sup_ref = self.support_reference_map.get(supports_id)
                            if(sup_ref is not None):
                                self.support_reference_map[supports_id] += 1
            elif 'functionTerms' in aspect:
                for function_term in aspect['functionTerms']:
                    self.function_term_map[function_term["po"]] = function_term
            elif 'reifiedEdges' in aspect:
                for reified_edge in aspect["reifiedEdges"]:
                    self.reified_edges [reified_edge['node']] = reified_edge
            elif 'visualProperties' in aspect:
                # remove all references to view id
                for visual_properties in aspect["visualProperties"]:
                    visual_properties.pop('view', None)
                self.unclassified_cx.append(aspect)
            elif 'cyVisualProperties' in aspect:
                # remove all references to view id
                for visual_properties in aspect["cyVisualProperties"]:
                    visual_properties.pop('view', None)
                self.unclassified_cx.append(aspect)
            else:
                self.unclassified_cx.append(aspect)
        print ''

    def networkx_to_NdexGraph(networkx_G):
        """Converts a NetworkX into a NdexGraph object"""
        warnings.warn(deprecation_message)
        G = NdexGraph()
        node_id = 0
        node_dict = {}
        G.max_edge_id = 0
        for node_name, node_attr in networkx_G.nodes_iter(data=True):
            if node_attr.has_key('name'):
                G.add_node(node_id, node_attr)
            else:
                G.add_node(node_id, node_attr, name=node_name)
            node_dict[node_name] = node_id
            node_id += 1
        for s, t, edge_attr in networkx_G.edges_iter(data=True):
            G.add_edge(node_dict[s], node_dict[t], G.max_edge_id, edge_attr)
            G.max_edge_id += 1

        if hasattr(networkx_G, 'pos'):
            G.pos = {node_dict[a] : b for a, b in networkx_G.pos.items()}
            #G.subnetwork_id = 1
            #G.view_id = 1

        return G


    def create_from_aspects(self, aspect, aspect_type):
        """ adds the corresponding networkn properties from a self-contained aspect.

        :param aspect: json based aspect
        :type aspect: dict
        :param aspect_type: name of aspect
        :type aspect_type: str
        :return: None
        :rtype: None
        """
        warnings.warn(deprecation_message)
        self.pos = {}
        #self.unclassified_cx = []

        if 'subNetworks' == aspect_type:
            for subnetwork in aspect:
                id = subnetwork['@id']
                if self.subnetwork_id != None:
                    raise ValueError("networkn does not support collections!")
                self.subnetwork_id = id

        elif 'cyViews' == aspect_type:
            for cyViews in aspect:
                id = cyViews['@id']
                if self.view_id != None:
                    raise ValueError("networkn does not support more than one view!")
                self.view_id = id

        elif 'metaData' == aspect_type:
            this_type_is_ignored_for_now = 'placeholder'

        elif 'nodes' == aspect_type:
            for node in aspect:
                id = node['@id']
                name = node['n'] if 'n' in node else None
                if name:
                    self.add_node(id, name=name)
                else:
                    self.add_node(id)
                represents = node['r'] if 'r' in node else None
                if represents:
                    self.node[id]['represents'] = represents

        elif 'edges' == aspect_type:
            for edge in aspect:
                id = edge['@id']
                interaction = edge['i'] if 'i' in edge else None
                s = edge['s']
                t = edge['t']
                self.edgemap[id] = (s, t)
                if interaction:
                    self.add_edge(s, t, key=id, interaction=interaction)
                else:
                    self.add_edge(s, t, key=id)

        elif 'networkAttributes' == aspect_type:
            for networkAttribute in aspect:
                name = networkAttribute['n']
                # special: ignore selected
                if name == 'selected':
                    continue
                value = networkAttribute['v']
                if 'd' in networkAttribute:
                    d = networkAttribute['d']
                    if d == 'boolean':
                        value = value.lower() == 'true'
                if 's' in networkAttribute or name not in self.graph:
                    self.graph[name] = value

        elif 'nodeAttributes' == aspect_type:
            for nodeAttribute in aspect:
                id = nodeAttribute['po']
                name = nodeAttribute['n']
                # special: ignore selected
                if name == 'selected':
                    continue
                value = nodeAttribute['v']
                if 'd' in nodeAttribute:
                    d = nodeAttribute['d']
                    if d == 'boolean':
                        value = value.lower() == 'true'
                if 's' in nodeAttribute or name not in self.node[id]:
                    self.node[id][name] = value

        elif 'edgeAttributes' == aspect_type:
            for edgeAttribute in aspect:
                id = edgeAttribute['po']
                s, t = self.edgemap[id]
                name = edgeAttribute['n']
                # special: ignore selected and shared_name columns
                if name == 'selected' or name == 'shared name':
                    continue
                value = edgeAttribute['v']
                if 'd' in edgeAttribute:
                    d = edgeAttribute['d']
                    if d == 'boolean':
                        value = value.lower() == 'true'
                if 's' in edgeAttribute or name not in self[s][t][id]:
                    self[s][t][id][name] = value

        elif 'cartesianLayout' == aspect_type:
            for nodeLayout in aspect:
                id = nodeLayout['node']
                x = nodeLayout['x']
                y = nodeLayout['y']
                self.pos[id] = [x,y]

        else:
            self.unclassified_cx.append({aspect_type: aspect})
            #self.unclassified_cx.append(aspect)

    def create_from_edge_list(self, edge_list, interaction='interacts with'):
        """ Create a network from a list of tuples that represent edges. Each tuple consists of  two nodes names that are to be connected. This operation will first clear ALL data already in the network.

            :param edge_list: A list of tuples containing node names that are connected.
            :type edge_list: list
            :param interaction: Either a list of interactions that is the same length as the edge_list a single string with an interaction to be applied to all edges.
            :type interaction: str or list

        """
        warnings.warn(deprecation_message)
        self.clear()
        node_names_added = {}
        for i, edge in enumerate(edge_list):
            #name 1 and name 2 are the node names from the tuple.
            name1 = edge[0]
            name2 = edge[1]
            if name1 in node_names_added:
                n1 = node_names_added[name1]
            else:
                n1 = self.add_new_node(name1)
                node_names_added[name1] = n1
            if name2 in node_names_added:
                n2 = node_names_added[name2]
            else:
                n2 = self.add_new_node(name2)
                node_names_added[name2] = n2

            edge_interaction = interaction if isinstance(interaction,string_types) else interaction[i]
            self.add_edge_between(n1,n2,edge_interaction)

    #------------------------------------------
    #       NETWORK
    #------------------------------------------

    def clear(self):
        """Eliminate all graph data in this network.  The network will then be empty and can be filled with data starting "from scratch".


        """
        warnings.warn(deprecation_message)
        super(NdexGraph, self).clear()
        self.subnetwork_id = None
        self.view_id = None
        self.max_node_id = None
        self.max_edge_id = None
        self.max_citation_id = None
        self.max_support_id = None
        self.pos = None
        self.edgemap = {}
        self.unclassified_cx = []
        self.citation_map = {}
        self.node_citation_map = {}
        self.edge_citation_map = {}
        self.support_map = {}
        self.node_support_map = {}
        self.edge_support_map = {}

    def set_name(self, name):
        """Set the name of this graph

        :param name: A descriptive name for the network which will show up on NDEx.
        :type name: str

        """
        warnings.warn(deprecation_message)
        self.graph['name'] = name

    def set_network_attribute(self, name, value):
    #    if name in NDEXGRAPH_RESERVED_ATTRIBUTES:
    #        raise ValueError(str(name) + " is a reserved network attribute name and may not be set by this method")
        warnings.warn(deprecation_message)
        self.graph[name] = value

    def get_name(self):
        """Get the name of this graph

        :return: The descriptive name for this network.
        :rtype: str

        """
        warnings.warn(deprecation_message)
        if 'name' in self.graph:
            return self.graph['name']
        return None

    def set_namespace (self, namespaceMap):
        warnings.warn(deprecation_message)
        self.namespaces = namespaceMap

    def show_stats(self):
        """Show the number of nodes and edges."""
        warnings.warn(deprecation_message)
        print("Nodes: %d" % self.number_of_nodes())
        print("Edges: %d" % self.number_of_edges())

    def set_edgemap(self, edgemap):
        warnings.warn(deprecation_message)
        self.edgemap = edgemap

    def subgraph_new(self, nbunch):
        warnings.warn(deprecation_message)

        return_graph = self.subgraph(nbunch[:400])

        for s, t, edge_id, data in return_graph.edges_iter(keys=True, data=True):
            return_graph.edgemap[edge_id] = (s, t)

        return return_graph

    #------------------------------------------
    #       OUTPUT
    #------------------------------------------

    def to_cx(self, md_dict=None):
        """Convert this network to a CX dictionary

        :return: The cx dictionary that represents this network.
        :rtype: dict
        """
        warnings.warn(deprecation_message)
        self.subnetwork_id = None
        self.view_id = None

        has_single_subnetwork = False
        if self.subnetwork_id and self.view_id:
            has_single_subnetwork = True
        if (self.subnetwork_id and not self.view_id) or (not self.subnetwork_id and self.view_id):
            raise ValueError("subnetwork and view inconsistent. subnetwork id = %s and view id = %s" % (self.subnetwork_id, self.view_id))

        G = self
        cx = []
        cx += ca.number_verification()
        cx += self.generate_metadata(G, self.unclassified_cx) #ca.metadata(metadata_dict=md_dict)

        #always add context first.
        if self.namespaces:
            cx += ca.namespaces(G)

        network_attributes = ca.network_attributes(G, has_single_subnetwork)
        cx += network_attributes

        if has_single_subnetwork:
            cx += ca.subnetworks(G, self.subnetwork_id, self.view_id)
        # - don't output subnetworks if the NdexGraph doesn't know about them.
        # - All operations that add aspects for visual properties, cartesian coordinates,
        #   or otherwise refer to subnetworks must ensure that subnetwork and view ids are set
        # else:
        #     cx += ca.subnetworks(G, 0, 0)
        cx += ca.nodes(G)
        cx += ca.edges(G)
        node_att = ca.node_attributes(G, has_single_subnetwork)
        if(node_att is not None):
            cx += node_att

        edge_att = ca.edge_attributes(G, has_single_subnetwork)
        if(edge_att is not None):
            cx += edge_att

        if self.pos and len(self.pos):
            if has_single_subnetwork:
                cx += ca.cartesian(G)#, self.view_id)
            else:
                cx += ca.cartesian(G)
                #raise ValueError("NdexGraph positions (g.pos) set without setting view and subnetwork ids")

        if len(self.citation_map) > 0:
            cx += ca.citations(G)
        if len(self.node_citation_map) > 0:
            cx += ca.node_citations(G)
        if len(self.edge_citation_map) > 0:
            cx += ca.edge_citations(G)
        if len(self.support_map) > 0:
            cx += ca.supports(G)
        if len(self.node_support_map) > 0:
            cx += ca.node_supports(G)
        if len(self.edge_support_map) > 0:
            cx += ca.edge_supports(G)
        if len(self.function_term_map) > 0:
            cx += ca.function_terms(G)
        if len(self.reified_edges) > 0:
            cx += ca.reified_edges(G)
        if self.provenance:
            cx += ca.provenance(G)

        for fragment in self.unclassified_cx:
            # filter out redundant networkRelations
            if not "networkRelations" in fragment:
                cx += [fragment]

        #cx += self.unclassified_cx
        cx += [self.status]

        return cx

    # Obsolete?
    def add_status(self, status):
        warnings.warn(deprecation_message)
        if(len(self.status['status']) == 0):
            self.status['status'].append(status)

    def generate_metadata(self, G, unclassified_cx):
        warnings.warn(deprecation_message)
        return_metadata = []
        consistency_group = 1
        if(self.metadata_original is not None):
            for mi in self.metadata_original:
                if(mi.get("consistencyGroup") is not None):
                    if(mi.get("consistencyGroup") > consistency_group):
                        consistency_group = mi.get("consistencyGroup")

            consistency_group += 1 # bump the consistency group up by one

            print("consistency group max: " + str(consistency_group))

        # ========================
        # @context metadata
        # ========================
        if  self.namespaces:
            return_metadata.append(
                {
                    "consistencyGroup": consistency_group,
                    "elementCount": 1,
                    "name": "@context",
                    "properties": [],
                    "version": "1.0"
                }
            )

        #========================
        # Nodes metadata
        #========================
        node_ids = [n[0] for n in G.nodes_iter(data=True)]
        if(len(node_ids) < 1):
            node_ids = [0]
        return_metadata.append(
            {
                "consistencyGroup" : consistency_group,
                "elementCount" : len(node_ids),
                "idCounter": max(node_ids),
                "name" : "nodes",
                "properties" : [ ],
                "version" : "1.0"
            }
        )

        #========================
        # Edges metadata
        #========================
        edge_ids = [e[2]for e in G.edges_iter(data=True, keys=True)]
        if(len(edge_ids) < 1):
            edge_ids = [0]
        return_metadata.append(
            {
                "consistencyGroup" : consistency_group,
                "elementCount" : len(edge_ids),
                "idCounter": max(edge_ids),
                "name" : "edges",
                "properties" : [ ],
                "version" : "1.0"
            }
        )

        #=============================
        # Network Attributes metadata
        #=============================
        if(len(G.graph) > 0):
            return_metadata.append(
                {
                    "consistencyGroup" : consistency_group,
                    "elementCount" : len(G.graph),
                    "name" : "networkAttributes",
                    "properties" : [ ],
                    "version" : "1.0"
                }
            )

        #===========================
        # Node Attributes metadata
        #===========================
        #id_max = 0
        attr_count = 0
        for node_id , attributes in G.nodes_iter(data=True):
            for attribute_name in attributes:
                if attribute_name != "name" and attribute_name != "represents":
                    attr_count += 1



        #
        # for n, nattr in G.nodes(data=True):
        #     if(bool(nattr)):
        #         attr_count += len(nattr.keys())
        #
        #     if(n > id_max):
        #         id_max = n

        if(attr_count > 0):
            return_metadata.append(
                {
                    "consistencyGroup" : consistency_group,
                    "elementCount" : attr_count,
                    #"idCounter": id_max,
                    "name" : "nodeAttributes",
                    "properties" : [ ],
                    "version" : "1.0"
                }
            )

        #===========================
        # Edge Attributes metadata
        #===========================
        #id_max = 0
        attr_count = 0
        for s, t, id, a in G.edges(data=True, keys=True):
            if(bool(a)):
                for attribute_name in a:
                    if attribute_name != "interaction":
                        attr_count += 1
                #attr_count += len(a.keys())

            # if(id > id_max):
            #     id_max = id

        if(attr_count > 0):
            return_metadata.append(
                {
                    "consistencyGroup" : consistency_group,
                    "elementCount" : attr_count,
                    #"idCounter": id_max,
                    "name" : "edgeAttributes",
                    "properties" : [ ],
                    "version" : "1.0"
                }
            )

        #===========================
        # cyViews metadata
        #===========================
        if self.view_id != None:
            return_metadata.append(
                {
                    "elementCount": 1,
                    "name": "cyViews",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )

        #===========================
        # subNetworks metadata
        #===========================
        if self.subnetwork_id != None:
            return_metadata.append(
                {
                    "elementCount": 1,
                    "name": "subNetworks",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )

        #===========================
        # networkRelations metadata
        #===========================
        if self.subnetwork_id != None and self.view_id != None:
            return_metadata.append(
                {
                    "elementCount": 2,
                    "name": "networkRelations",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )

        #===========================
        # citations and supports metadata
        #===========================
        if len(self.support_map) > 0:
            return_metadata.append(
                {
                    "elementCount": len(self.support_map),
                    "name": "supports",
                    "properties": [],
                    "idCounter": max(self.support_map.keys()),
                    "consistencyGroup" : consistency_group
                }
            )

        if len(self.node_support_map) > 0:
            return_metadata.append(
                {
                    "elementCount": len(self.node_support_map),
                    "name": "nodeSupports",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )
        if len(self.edge_support_map) > 0:
            return_metadata.append(
                {
                    "elementCount": len(self.edge_support_map),
                    "name": "edgeSupports",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )

        if len(self.citation_map) > 0:
            return_metadata.append(
                {
                    "elementCount": len(self.citation_map),
                    "name": "citations",
                    "properties": [],
                    "idCounter": max(self.citation_map.keys()),
                    "consistencyGroup" : consistency_group
                }
            )

        if len(self.node_citation_map) > 0:
            return_metadata.append(
                {
                    "elementCount": len(self.node_citation_map),
                    "name": "nodeCitations",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )

        if len(self.edge_citation_map) > 0:
            return_metadata.append(
                {
                    "elementCount": len(self.edge_citation_map),
                    "name": "edgeCitations",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )

        if len(self.function_term_map) > 0:
            return_metadata.append(
                {
                    "elementCount": len(self.function_term_map),
                    "name": "functionTerms",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )

        if len(self.reified_edges) > 0:
            return_metadata.append(
                {
                    "elementCount": len(self.reified_edges),
                    "name": "reifiedEdges",
                    "properties": [],
                    "consistencyGroup" : consistency_group
                }
            )

        #===========================
        # ndexStatus metadata
        #===========================
        return_metadata.append(
            {
                "consistencyGroup": consistency_group,
                "elementCount": 1,
                "name": "ndexStatus",
                "properties": [],
                "version": "1.0"
            }
        )

        #===========================
        # cartesianLayout metadata
        #===========================
        if self.pos and len(self.pos) > 0:
            return_metadata.append(
                {
                    "consistencyGroup": consistency_group,
                    "elementCount": len(self.pos),
                    "name": "cartesianLayout",
                    "properties": [],
                    "version": "1.0"
                }
            )

        #===========================
        # OTHER metadata
        #===========================
        for asp in self.unclassified_cx:
            try:
                aspect_type = asp.iterkeys().next()
                if(aspect_type == "visualProperties"
                   or aspect_type == "cyVisualProperties"
                   or aspect_type == "@context"):
                    return_metadata.append(
                        {
                            "consistencyGroup" : consistency_group,
                            "elementCount":len(asp[aspect_type]),
                            "name":aspect_type,
                            "properties":[]
                         }
                    )
            except Exception as e:
                print(e.message)


        #print {'metaData': return_metadata}

        return [{'metaData': return_metadata}]


    def to_cx_stream(self, md_dict=None):
        """Convert this network to a CX stream

        :return: The CX stream representation of this network.
        :rtype: io.BytesIO

        """
        warnings.warn(deprecation_message)
        cx = self.to_cx(md_dict)

        if sys.version_info.major == 3:
            return io.BytesIO(json.dumps(cx).encode('utf-8'))
        else:
            return_bytes = None
            try:
                return_bytes = io.BytesIO(json.dumps(cx))
            except UnicodeDecodeError as err1:
                print "Detected invalid encoding. Trying latin-1 encoding."
                return_bytes = io.BytesIO(json.dumps(cx, encoding="latin-1"))
                print "Success"
            except Exception as err2:
                print err2.message

            return return_bytes

    def write_to(self, filename):
        """Write this network as a CX file to the specified filename.

        :param filename: The name of the file to write to.
        :type filename: str

        """
        warnings.warn(deprecation_message)
        with open(filename, 'w') as outfile:
            json.dump(self.to_cx(), outfile, indent=4)

    def upload_to(self, server, username, password):
        """ Upload this network to the specified server to the account specified by username and password.

        :param server: The NDEx server to upload the network to.
        :type server: str
        :param username: The username of the account to store the network.
        :type username: str
        :param password: The password for the account.
        :type password: str
        :return: The UUID of the network on NDEx.
        :rtype: str

        Example:
            ndexGraph.upload_to('http://test.ndexbio.org', 'myusername', 'mypassword')
        """

        warnings.warn(deprecation_message)
        ndex = nc.Ndex2(server,username,password)
        return ndex.save_new_network(self.to_cx())

    def update_to(self, uuid, server, username, password):
        """ Upload this network to the specified server to the account specified by username and password.

        :param server: The NDEx server to upload the network to.
        :type server: str
        :param username: The username of the account to store the network.
        :type username: str
        :param password: The password for the account.
        :type password: str
        :return: The UUID of the network on NDEx.
        :rtype: str

        Example:
            ndexGraph.upload_to('http://test.ndexbio.org', 'myusername', 'mypassword')
        """
        warnings.warn(deprecation_message)
        cx = self.to_cx()
        ndex = nc.Ndex2(server,username,password)

        if(len(cx) > 0):
            if(cx[len(cx) - 1] is not None):
                if(cx[len(cx) - 1].get('status') is None):
                    # No STATUS element in the array.  Append a new status
                    cx.append({"status" : [ {"error" : "","success" : True} ]})
                else:
                    if(len(cx[len(cx) - 1].get('status')) < 1):
                        # STATUS element found, but the status was empty
                        cx[len(cx) - 1].get('status').append({"error" : "","success" : True})

            if sys.version_info.major == 3:
                stream = io.BytesIO(json.dumps(cx).encode('utf-8'))
            else:
                stream = io.BytesIO(json.dumps(cx))

            return ndex.update_cx_network(stream, uuid)
        else:
            raise IndexError("Cannot save empty CX.  Please provide a non-empty CX document.")

    #------------------------------------------
    #       NODES
    #------------------------------------------

    def add_new_node(self, name=None, attr_dict=None, **attr):
        """Add a cx node, possibly with a particular name, to the graph.

        :param name: The name of the node. (Optional).
        :type name: str
        :param attr_dict: Dictionary of node attributes.  Key/value pairs will update existing data associated with the node.
        :type attr_dict: dict
        :param attr: Set or change attributes using key=value.

        """
        warnings.warn(deprecation_message)
        if self.max_node_id == None:
            if self.number_of_nodes() > 0:
                self.max_node_id = max(self.nodes())
            else:
                self.max_node_id = 0
        self.max_node_id += 1
        if name:
            attr['name'] = name
        self.add_node(self.max_node_id, attr_dict, **attr)
        return self.max_node_id

    def add_cx_node(self, id, name=None, represents = None, attr_dict=None):

        warnings.warn(deprecation_message)
        if (not attr_dict) and (name or represents):
            attr_dict = {}
        if name:
            attr_dict['name'] = name
        if represents:
            attr_dict['represents'] = represents
        self.add_node(id, attr_dict)

    def get_cx_node (self, id):
        warnings.warn(deprecation_message)
        return self.node[id]

    def remove_nodes_from(self, nbunch):
        warnings.warn(deprecation_message)
        for n in nbunch:
            self.pos.pop(n, None)
            self.function_term_map.pop(n, None)
        super(MultiDiGraph, self).remove_nodes_from(nbunch)

    def remove_node(self, n):
        warnings.warn(deprecation_message)
        self.reified_edges.pop(n,None)

        if(self.function_term_map.get(n) is not None):
            self.function_term_map.pop(n, None)

        self.remove_citation_and_support_node_references(n)

        super(MultiDiGraph, self).remove_node(n)

    def remove_citation_and_support_node_references(self, node_id):

        # remove support to edge references
        warnings.warn(deprecation_message)
        if self.node_support_map.has_key(node_id):
            # get the supports that reference the edge
            support_ids = self.node_support_map[node_id]
            # remove the edge entry from the edge_support_map

            #=====================================================
            # Check the "supports" reference map. Decrement the
            # reference value and if it reaches 0 remove from map
            #=====================================================
            supports_array = self.node_support_map.get(node_id)
            for supports_array_id in supports_array:
                decrement_this = self.support_reference_map.get(supports_array_id)
                if(decrement_this is not None):
                    decrement_this -= 1
                    self.support_reference_map[supports_array_id] = decrement_this
                    if(decrement_this == 0):
                        self.support_map.pop(supports_array_id)
                        self.support_reference_map.pop(supports_array_id)

            self.node_support_map.pop(node_id, None)



        # remove support to edge references
        if self.node_citation_map.has_key(node_id):
            #=====================================================
            # Check the "citations" reference map. Decrement the
            # reference value and if it reaches 0 remove from map
            #=====================================================
            citations_array = self.node_citation_map.get(node_id)
            for citations_array_id in citations_array:
                decrement_this = self.citation_reference_map.get(citations_array_id)
                if(decrement_this is not None):
                    decrement_this -= 1
                    self.citation_reference_map[citations_array_id] = decrement_this
                    if(decrement_this == 0):
                        self.citation_map.pop(citations_array_id)
                        self.citation_reference_map.pop(citations_array_id)

            self.node_citation_map.pop(node_id, None)

    def remove_orphan_nodes(self):
        #   remove nodes with no edges
        warnings.warn(deprecation_message)
        for node_id in self.nodes():
            #node_name = network.get_node_attribute_value_by_id(node_id)
            degree = self.degree([node_id])[node_id]
            #print node_name + " : " + str()
            if degree is 0:
                #print " -- removing " + str(node_name) + " " + str(node_id)
                self.remove_node(node_id)

    def get_node_ids(self, value, query_key='name'):
        """Returns a list of node ids of all nodes in which query_key has the specified value.

            :param value: The value we want.
            :param query_key: The name of the attribute where we should look for the value.
            :return: A list of node ids.
            :rtype: list

        """
        warnings.warn(deprecation_message)
        nodes = [n[0] for n in self.nodes_iter(data=True) if query_key in n[1] and n[1][query_key] == value]
        return nodes


    def set_node_attribute(self, id, attribute_key, attribute_value):
        """Set the value of a particular edge attribute.

            :param id: The edge id we wish to set an attribute on.
            :type id: int
            :param attribute_key: The name of the attribute we wish to set.
            :type attribute_key: string
            :param attribute_value: The value we wish to set the attribute to.
            :type attribute_value: any

        """
        warnings.warn(deprecation_message)
        self.node[id][attribute_key] = attribute_value

    def get_node_attribute_value_by_id(self, node_id, query_key='name', error=False):
        """Get the value of a particular node attribute based on the id.

        :param node_id: A node id.
        :type node_id: int
        :param query_key: The name of the attribute whose value we desire.
        :type query_key: str
        :return: The attribute value.
        :rtype: any

        """
        warnings.warn(deprecation_message)
        if node_id not in self.node:
            raise ValueError("Your ID is not in the network")
        node_attributes = nx.get_node_attributes(self, query_key)
        if error and len(node_attributes) == 0:
            raise ValueError("That node attribute name does not exist ANYWHERE in the network.")
        return self.node[node_id][query_key] if query_key in self.node[node_id] else None

    def get_node_attribute_values_by_id_list(self, id_list, query_key='name'):
        """Returns a list of attribute values that correspond with the attribute key using the nodes in id_list.

            :param id_list: A list of node ids whose attribute values we wish to retrieve.
            :param query_key: The name of the attribute whose corresponding values should be retrieved.
            :return: A list of attribute values.
            :rtype: list

        """
        warnings.warn(deprecation_message)
        for id in id_list:
            if id not in self.node:
                raise ValueError("Your ID list has IDs that are not in the network")
        node_attributes = nx.get_node_attributes(self, query_key)
        if len(node_attributes) == 0:
            raise ValueError("That node attribute name does not exist ANYWHERE in the network.")
        return [self.node[id][query_key] if query_key in self.node[id] else None for id in id_list]

    def get_node_names_by_id_list(self, id_list):
        """Given a list of node ids, return a list of node names.

            :param id_list: A list of node ids whose attribute values we wish to retrieve.
            :type id_list: list
            :return: A list of node names.
            :rtype: list

        """
        warnings.warn(deprecation_message)
        return self.get_node_attribute_values_by_id_list(id_list)

    def get_node_name_by_id(self, id):
        """Given a node id, return the name of the node.

        :param id: The cx id of the node.
        :type id: int
        :return: The name of the node.
        :rtype: str

        """
        warnings.warn(deprecation_message)
        return self.get_node_attribute_value_by_id(id)

    def get_all_node_attribute_keys(self):
        """Get the unique list of all attribute keys used in at least one node in the network.

            :return: A list of attribute keys.
            :rtype: list

        """
        warnings.warn(deprecation_message)
        keys = set()
        for _, attributes in self.nodes_iter(data=True):
            for key, value in attributes.iteritems():
                keys.add(key)
        return list(keys)


    #------------------------------------------
    #       Edges
    #------------------------------------------

    def get_edge_ids_by_node_attribute(self, source_node_value, target_node_value, attribute_key='name'):
        """Returns a list of edge ids of all edges where both the source node and target node have the specified values for attribute_key.

                :param source_node_value: The value we want in the source node.
                :param target_node_value: The value we want in the target node.
                :param attribute_key: The name of the attribute where we should look for the value.
                :return: A list of edge ids.
                :rtype: list
            """
        warnings.warn(deprecation_message)
        source_node_ids = self.get_node_ids(source_node_value, attribute_key)
        target_node_ids = self.get_node_ids(target_node_value, attribute_key)
        edge_keys = []
        for s in source_node_ids:
            for t in target_node_ids:
                if s in self and t in self[s]:
                    edge_keys += self[s][t].keys()
        return edge_keys

    def add_edge_between(self, source_node_id, target_node_id, interaction='interacts_with', attr_dict=None, **attr):
        """Add an edge between two nodes in this network specified by source_node_id and target_node_id, optionally specifying the type of interaction.

            :param source_node_id: The node id of the source node.
            :type source_node_id: int
            :param target_node_id: The node id of the target node.
            :type target_node_id: int
            :param interaction: The type of interaction specified by the newly added edge.
            :type interaction: str
            :param attr_dict: Dictionary of node attributes.  Key/value pairs will update existing data associated with the node.
            :type attr_dict: dict
            :param attr: Set or change attributes using key=value.
        """
        warnings.warn(deprecation_message)
        if source_node_id not in self.node:
            raise ValueError('source_node_id = %d is not in the network' % source_node_id)
        if target_node_id not in self.node:
            raise ValueError('target_node_id = %d is not in the network' % target_node_id)

        if self.max_edge_id == None:
            if self.number_of_edges() > 0:
                self.max_edge_id = max([x[2] for x in self.edges(keys=True)])
            else:
                self.max_edge_id = 0
        self.max_edge_id += 1
        self.add_edge(source_node_id, target_node_id, self.max_edge_id, interaction=interaction, attr_dict=attr_dict, **attr)
        self.edgemap[self.max_edge_id] = (source_node_id, target_node_id)
        return self.max_edge_id

    def get_node_ids_by_edge_id(self, edge_id):
        warnings.warn(deprecation_message)
        if edge_id in self.edgemap:
            s, t = self.edgemap[edge_id]
            return s, t
        else:
            raise Exception("edge id " + str(edge_id) + " not found in network")
        
    def remove_edge_by_id(self, edge_id):
        warnings.warn(deprecation_message)
        source_id, target_id = self.get_node_ids_by_edge_id(edge_id)

        # remove edge from edge map
        self.edgemap.pop(edge_id, None)
        pop_these_reified_edges = []

        for n,re in self.reified_edges.iteritems():
            if(re["edge"] == edge_id):
                pop_these_reified_edges.append(n)
                # This causes problems when editing the dictionary while iterating over it --> self.reified_edges.pop(n,None)

        for n in pop_these_reified_edges:
            self.reified_edges.pop(n,None)

        #self.edge_citation_map.pop(edge_id, None)
        #self.edge_support_map.pop(edge_id, None)

        # remove citation and support references to edges
        self.remove_citation_and_support_edge_references(edge_id)

        # networkX remove edge
        self.remove_edge(source_id, target_id, edge_id)

    def remove_citation_and_support_edge_references(self, edge_id):
        warnings.warn(deprecation_message)
        citation_ids = None
        support_ids = None

        # remove support to edge references
        if self.edge_support_map.has_key(edge_id):
            # get the supports that reference the edge
            support_ids = self.edge_support_map[edge_id]
            # remove the edge entry from the edge_support_map

            #=====================================================
            # Check the "supports" reference map. Decrement the
            # reference value and if it reaches 0 remove from map
            #=====================================================
            supports_array = self.edge_support_map.get(edge_id)
            for supports_array_id in supports_array:
                decrement_this = self.support_reference_map.get(supports_array_id)
                if(decrement_this is not None):
                    decrement_this -= 1
                    self.support_reference_map[supports_array_id] = decrement_this
                    if(decrement_this == 0):
                        self.support_map.pop(supports_array_id)
                        self.support_reference_map.pop(supports_array_id)

            self.edge_support_map.pop(edge_id)



        # remove support to edge references
        if self.edge_citation_map.has_key(edge_id):
            #=====================================================
            # Check the "citations" reference map. Decrement the
            # reference value and if it reaches 0 remove from map
            #=====================================================
            citations_array = self.edge_citation_map.get(edge_id)
            for citations_array_id in citations_array:
                decrement_this = self.citation_reference_map.get(citations_array_id)
                if(decrement_this is not None):
                    decrement_this -= 1
                    self.citation_reference_map[citations_array_id] = decrement_this
                    if(decrement_this == 0):
                        self.citation_map.pop(citations_array_id)
                        self.citation_reference_map.pop(citations_array_id)

            self.edge_citation_map.pop(edge_id)


            '''
            # eliminate the supports that are still referenced by some node or edge
            for map_support_ids in self.edge_support_map.values():
                for map_support_id in map_support_ids:
                    if map_support_id in support_ids:
                        support_ids.remove(map_support_id)
                        if len(support_ids) == 0:
                            # we have proved that all the potentially orphaned
                            # supports are connected to some other edge
                            break

            for map_support_ids in self.node_support_map.values():
                for map_support_id in map_support_ids:
                    if map_support_id in support_ids:
                        support_ids.remove(map_support_id)
                        if len(support_ids) == 0:
                            # we have proved that all the potentially orphaned
                            # supports are connected to some node
                            break

            # the remaining supports are orphaned and may be removed
            for support_id in support_ids:
                self.support_map.pop(support_id)


        # remove citation to edge references
        if edge_id in self.edge_citation_map:
            # get the citations that reference the edge
            citation_ids = self.edge_citation_map[edge_id]
            # remove the edge entry from the edge_citation_map
            self.edge_citation_map.pop(edge_id)

            # eliminate the citations that are still referenced by some node or edge
            for map_citation_ids in self.edge_citation_map.values():
                for map_citation_id in map_citation_ids:
                    if map_citation_id in citation_ids:
                        citation_ids.remove(map_citation_id)
                        if len(citation_ids) == 0:
                            # we have proved that all the potentially orphaned
                            # citations are connected to some other edge
                            break

            for map_citation_ids in self.node_citation_map.values():
                for map_citation_id in map_citation_ids:
                    if map_citation_id in citation_ids:
                        citation_ids.remove(map_citation_id)
                        if len(citation_ids) == 0:
                            # we have proved that all the potentially orphaned
                            # citations are connected to some node
                            break

            # the remaining citations are orphaned and may be removed
            for citation_id in citation_ids:
                self.citation_map.pop(citation_id)
        '''
        # at this point, all supports that referenced a removed citation
        # have already been removed.


    #TODO Check args
    def get_edge_attribute_value_by_id(self, edge_id, attribute_key):
        """Get the value for attribute of the edge specified by edge_id.

            :param edge_id: The id of the edge.
            :param query_key: The name of the attribute whose value should be retrieved.
            :return: The attribute value. If the value is a list, this returns the entire list.
            :rtype: any

        """
        # edge_keys = {key: (s, t) for s, t, key in self.edges_iter(keys=True)}
        # if edge_id not in edge_keys:
        #     raise ValueError("edge ID Your ID is not in the network")
        # s = edge_keys[edge_id][0]
        # t = edge_keys[edge_id][1]
        # edge_attributes = nx.get_edge_attributes(self, query_key)
        # if len(edge_attributes) == 0:
        #     raise ValueError("That node edge name does not exist ANYWHERE in the network.")
        warnings.warn(deprecation_message)
        source_id, target_id = self.get_node_ids_by_edge_id(edge_id)
        return self[source_id][target_id][edge_id][attribute_key] if attribute_key in self[source_id][target_id][edge_id] else None

    #TODO Check args
    def set_edge_attribute(self, edge_id, attribute_key, attribute_value):
        """Set the value of a particular edge attribute.

            :param edge_id: The edge id we wish to set an attribute on.
            :type edge_id: int
            :param attribute_key: The name of the attribute we wish to set.
            :type attribute_key: string
            :param attribute_value: The value we wish to set the attribute to.
            :type attribute_value: any

        """
        warnings.warn(deprecation_message)
        source_id, target_id = self.get_node_ids_by_edge_id(edge_id)
        self.edge[source_id][target_id][edge_id][attribute_key] = attribute_value

    def get_edge_ids_by_source_target(self, source_id, target_id):
        warnings.warn(deprecation_message)
        edge_ids =[]
        out_edges = self.out_edges(keys=True)
        for edge in out_edges:
            if edge[1] == target_id:
                edge_ids.append(edge[2])
        return edge_ids

    def get_edge_attribute_values_by_id_list(self, edge_id_list, attribute_key):
        """Given a list of edge ids and particular attribute key, return a list of corresponding attribute values.'

            :param edge_id_list: A list of edge ids whose attribute values we wish to retrieve
            :param query_key: The name of the attribute whose corresponding values should be retrieved.
            :return: A list of attribute values corresponding to the edge keys input.
            :rtype: list

        """
        warnings.warn(deprecation_message)
        for edge_id in edge_id_list:
            if edge_id not in self.edgemap:
                raise ValueError("edge id list error: edge id " + str(edge_id) + " not found in network")
            
        edge_keys = {id: self.edgemap[id] for id in edge_id_list}
        edge_attributes = nx.get_edge_attributes(self, attribute_key)
        if len(edge_attributes) == 0:
            raise ValueError("the attribute name " + str(attribute_key) + " is not used ANYWHERE in the network.")
        
        return [self[v[0]][v[1]][k][attribute_key] if attribute_key in self[v[0]][v[1]][k] else None for k, v in
                edge_keys.iteritems()]

    def get_all_edge_attribute_keys(self):
        """Get the unique list of all attribute keys used in at least one edge in the network.

        :return: A list of edge attribute keys.
        :rtype: list

        """
        warnings.warn(deprecation_message)
        keys = set()
        for _, _, attributes in self.edges_iter(data=True):
            for key, value in attributes.iteritems():
                keys.add(key)
        return list(keys)


    #------------------------------------------
    #       Citations
    #------------------------------------------

    def add_citation(self, identifier=None, title=None, description=None, attributes=None):
        # get the next citation id
        warnings.warn(deprecation_message)
        if self.max_citation_id == None:
            if len(self.citation_map) > 0:
                self.max_citation_id = max(self.citation_map.keys())
            else:
                self.max_citation_id = 0
        self.max_citation_id += 1

        citation = {}
        if identifier:
            citation["dc:identifier"] = identifier
        if title:
            citation["dc:title"] = title
        if description:
            citation["description"] = description
        if attributes:
            citation["attributes"] = copy.deepcopy(attributes)
        self.citation_map[self.max_citation_id] = citation
        return self.max_citation_id

    def add_node_citation_ref(self, node_id, citation_id):
        warnings.warn(deprecation_message)
        if node_id in self.node_citation_map:
            citation_ids = self.node_citation_map[node_id]
            if citation_id not in citation_ids:
                citation_ids.append(citation_id)
        else:
            citation_ids = [citation_id]
            self.node_citation_map[node_id] = citation_ids

    def add_citation_to_edge(self, edge_id, citation_string):
        #TODO check edge attriutes for ndex:citation.  Add list if not exist otherwise append.  No duplicates
        #self.set_edge_attribute(edge_id, 'ndex:citation', citation_string)

        warnings.warn(deprecation_message)
        source_id, target_id = self.get_node_ids_by_edge_id(edge_id)
        if self.edge[source_id][target_id][edge_id].get('ndex:citation') is None:
            self.edge[source_id][target_id][edge_id]['ndex:citation'] = [citation_string]
        else:
            self.edge[source_id][target_id][edge_id]['ndex:citation'].append(citation_string)


    def add_edge_citation_ref(self, edge_id, citation_id, override=None):
        warnings.warn(deprecation_message)
        if override is None:
            raise Exception("This method has been deprecated because it is incompatible with Cytoscape. "
                            "Use add_citation_to_edge() instead. You can still use this method if you set "
                            "override=True in the parameters.")

        if edge_id in self.edge_citation_map:
            citation_ids = self.edge_citation_map[edge_id]
            if citation_id not in citation_ids:
                citation_ids.append(citation_id)
        else:
            citation_ids = [citation_id]
            self.edge_citation_map[edge_id] = citation_ids
            
    #------------------------------------------
    #       Supports
    #------------------------------------------

    def add_support(self, citation=None, text=None, attributes=None):
        # get the next support id
        warnings.warn(deprecation_message)
        if self.max_support_id == None:
            if len(self.citation_map) > 0:
                self.max_support_id = max(self.support_map.keys())
            else:
                self.max_support_id = 0
        self.max_support_id += 1

        support = {}
        if citation:
            support["citation"] = citation
        if text:
            support["text"] = text
        if attributes:
            support["attributes"] = copy.deepcopy(attributes)
        self.support_map[self.max_support_id] = support
        return self.max_support_id
        
    def add_node_support_ref(self, node_id, support_id):
        warnings.warn(deprecation_message)
        if node_id in self.node_support_map:
            support_ids = self.node_support_map[node_id]
            if support_id not in support_ids:
                support_ids.append(support_id)
        else:
            support_ids = [support_id]
            self.node_support_map[node_id] = support_ids

    def add_edge_support_ref(self, edge_id, support_id):
        warnings.warn(deprecation_message)
        if edge_id in self.edge_support_map:
            support_ids = self.edge_support_map[edge_id]
            if support_id not in support_ids:
                support_ids.append(support_id)
        else:
            support_ids = [support_id]
            self.edge_support_map[edge_id] = support_ids

    #------------------------------------------
    #       Provenance
    #------------------------------------------

    def get_provenance(self):
        warnings.warn(deprecation_message)
        return self.provenance

    def set_provenance(self, provenance):
        warnings.warn(deprecation_message)
        self.provenance = provenance

    def update_provenance(self, event_type, entity_props=None):
        warnings.warn(deprecation_message)
        current_provenance = self.provenance
        if current_provenance:
            new_provenance = make_provenance(event_type, provenance=current_provenance, entity_props=entity_props)
        else:
            new_provenance = make_provenance(event_type,entity_props=entity_props)
        self.set_provenance(new_provenance)

    def set_namespaces(self,namespaces):
        warnings.warn(deprecation_message)
        self.namespaces = namespaces

    #------------------------------------------
    #       Datatypes
    #------------------------------------------


def make_provenance(event_type, provenance=None, entity_props=None):
    warnings.warn(deprecation_message)
    uri = None
    old_entity = None

    if provenance and "entity" in provenance:
        old_entity = provenance["entity"]
        if uri in old_entity:
            uri = old_entity["uri"]

    t = int(round(time() * 1000))

    event = {
        "startedAtTime": t,
        "endedAtTime": t,
        "eventType": event_type
    }

    if old_entity:
        event["inputs"] = [old_entity]

    entity = {"creationEvent": event}

    if uri:
        entity["uri"] = uri

    if entity_props:
        entity["properties"] = entity_props

    new_provenance = {"entity": entity}

    return new_provenance

class FilterSub:
    """A graph compatible with NDEx"""
    def __init__(self, cx=None, subnet_index=0, subnetwork_id=None):
        warnings.warn(deprecation_message)
        self.subnetwork_id = None
        self.view_id = None
        self.max_node_id = None
        self.max_edge_id = None
        self.max_citation_id = None
        self.max_support_id = None
        self.pos = {}
        self.unclassified_cx = []
        self.metadata_original = None
        self.status = {'status': [{'error' : '','success' : True}]}
        self.citation_map = {}
        self.node_citation_map = {}
        self.edge_citation_map = {}
        self.support_map = {}
        self.node_support_map = {}
        self.edge_support_map = {}
        self.function_term_map = {}
        self.cx = cx

        # Maps edge ids to node ids. e.g. { edge1: (source_node, target_node), edge2: (source_node, target_node) }
        self.edgemap = {}

        # If there is no CX to process raise exception
        if cx == None:
            raise RuntimeError("Missing CX. Please provide a valid CX")

        found_subnetwork = False
        for aspect in cx:
            if 'subNetworks' in aspect:
                found_subnetwork = True
                sub_nets = aspect.get('subNetworks')

                if(subnetwork_id is not None):
                    self.subnetwork_id = subnetwork_id
                    for s in sub_nets:
                        if(s.get("@id") == subnetwork_id):
                            sub_nets = [s]
                            break
                else:
                    try:
                        subnetwork = sub_nets[subnet_index]
                    except IndexError:
                        raise RuntimeError("Subnetwork at index %d does not exist " % subnet_index)

                    self.subnetwork_id = subnetwork.get('@id')
                    sub_nets = [subnetwork]
                    break

        if(not found_subnetwork):
            for aspect in cx:
                if 'cySubNetworks' in aspect:
                    found_subnetwork = True
                    sub_nets = aspect.get('cySubNetworks')
                    if(subnetwork_id is not None):
                        self.subnetwork_id = subnetwork_id
                        for s in sub_nets:
                            if(s.get("@id") == subnetwork_id):
                                sub_nets = [s]
                                break
                    else:
                        try:
                            subnetwork = sub_nets[subnet_index]
                        except IndexError:
                            raise RuntimeError("Subnetwork at index %d does not exist " % subnet_index)

                        self.subnetwork_id = subnetwork.get('@id')
                        sub_nets = [subnetwork]
                        break

        self.unclassified_cx = []
        for aspect in cx:
            if 'networkAttributes' in aspect:
                the_list = aspect['networkAttributes']
                the_list[:] = [d for d in the_list if d.get('s') == self.subnetwork_id]

            elif 'nodeAttributes' in aspect:
                the_list = aspect['nodeAttributes']
                the_list[:] = [d for d in the_list if d.get('s') == self.subnetwork_id]

            elif 'edgeAttributes' in aspect:
                the_list = aspect['edgeAttributes']
                the_list[:] = [d for d in the_list if d.get('s') == self.subnetwork_id]

            elif 'cyTableColumn' in aspect:
                the_list = aspect['cyTableColumn']
                the_list[:] = [d for d in the_list if d.get('s') == self.subnetwork_id]

            elif 'cyViews' in aspect:
                the_list = aspect['cyViews']
                the_list[:] = [d for d in the_list if d.get('s') == self.subnetwork_id]

            else:
                self.unclassified_cx.append(aspect)

        #print cx

    def get_cx(self):
        warnings.warn(deprecation_message)
        return self.cx


#!/usr/bin/env python

import requests
import json
import ndex2
from requests_toolbelt import MultipartEncoder
import os
import io
import sys
import decimal

if sys.version_info.major == 3:
    from urllib.parse import urljoin
else:
    from urlparse import urljoin

from requests import exceptions as req_except
import time
import base64

userAgent = 'NDEx-Python/3.0'

class Ndex2:


    '''A class to facilitate communication with an NDEx server.'''
    def __init__(self, host = "http://public.ndexbio.org", username = None, password = None, update_status=False, debug = False):
        '''Creates a connection to a particular NDEx server.

                :param host: The URL of the server.
                :type host: string
                :param username: The username of the NDEx account to use. (Optional)
                :type username: string
                :param password: The account password. (Optional)
                :type password: string
        '''
        self.debug = debug
        self.version = 1.3
        self.status = {}
        self.username = username
        self.password = password
        if "localhost" in host:
            self.host = "http://localhost:8080/ndexbio-rest"
        else:
            status_url = "/rest/admin/status"

            try:
                version_url = urljoin(host, status_url)


                response = requests.get(version_url,headers = {'User-Agent': userAgent})
                response.raise_for_status()
                data = response.json()

                prop = data.get('properties')
                if(prop is not None):
                    pv = prop.get('ServerVersion')
                    if(pv is not None):
                        if not pv.startswith('2.'):
                            raise Exception("This release only supports NDEx 2.x server.")
                        else:
                            self.version = pv
                            self.host = host + "/v2"
                    else:
                        sys.stderr.write("Warning: This release doesn't fully support 1.3 version of NDEx")
                        self.version = "1.3"
                        self.host = host + "/rest"

            except req_except.HTTPError as he:
                ndex2.get_logger('CLIENT').warning('Can''t determine server version.' + host + ' Server returned error -- '  + he.message)
                self.version = "1.3"
                self.host = host + "/rest"
                #TODO - how to handle errors getting server version...

        # create a session for this Ndex
        self.s = requests.session()
        if username and password:
            # add credentials to the session, if available
            self.s.auth = (username, password)

        if update_status:
            self.update_status()

# Base methods for making requests to this NDEx

    def set_debug_mode(self, debug):
        self.debug = debug

    def debug_response(self, response):
        if self.debug:
            print("status code: " + str(response.status_code))
            if not response.status_code == requests.codes.ok:
                print("response text: " + response.text)

    def require_auth(self):
        if not self.s.auth:
            raise Exception("this method requires user authentication")

    def put(self, route, put_json=None):
        url = self.host + route
        if self.debug:
            print("PUT route: " + url)
            if put_json is not None:
                print("PUT json: " + put_json)

        headers = self.s.headers
        headers['Content-Type'] ='application/json;charset=UTF-8'
        headers['Accept'] =  'application/json'
        headers['User-Agent']= userAgent

        if put_json is not None:
            response = self.s.put(url, data = put_json, headers = headers)
        else:
            response = self.s.put(url, headers = headers)
        self.debug_response(response)
        response.raise_for_status()
        if response.status_code == 204:
            return ""
        if response.headers['content-type'] == 'application/json':
            return response.json()
        else :
            return response.text


    def post(self, route, post_json):
        url = self.host + route
        if self.debug:
            print("POST route: " + url)
            print("POST json: " + post_json)
        headers = {'Content-Type': 'application/json',
                   'Accept': 'application/json,text/plain',
                   'Cache-Control': 'no-cache',
                   'User-Agent':  userAgent,
                   }
        response = self.s.post(url, data=post_json, headers=headers)
        self.debug_response(response)
        response.raise_for_status()
        if response.status_code == 204:
            return ""
        if response.headers['content-type'] == 'application/json':
            return response.json()
        else :
            return response.text

    def delete(self, route):
        url = self.host + route
        if self.debug:
            print("DELETE route: " + url)
        headers = self.s.headers
        headers['User-Agent'] = userAgent
        response = self.s.delete(url, headers = headers)
        self.debug_response(response)
        response.raise_for_status()
        if response.status_code == 204:
            return ""
        return response.json()

    def get(self, route, get_params = None):
        url = self.host + route
        if self.debug:
            print("GET route: " + url)
        headers = self.s.headers
        headers['User-Agent'] = userAgent
        response = self.s.get(url, params = get_params, headers=headers)
        self.debug_response(response)
        response.raise_for_status()
        if response.status_code == 204:
            return None
        return response.json()

    # The stream refers to the Response, not the Request
    def get_stream(self, route, get_params = None):
        url = self.host + route
        if self.debug:
            print("GET stream route: " + url)
        headers = self.s.headers
        headers['User-Agent'] = userAgent
        response = self.s.get(url, params = get_params, stream = True,headers = headers)
        self.debug_response(response)
        response.raise_for_status()
        if response.status_code == 204:
            return ""
        return response

    # The stream refers to the Response, not the Request
    def post_stream(self, route, post_json):
        url = self.host + route
        if self.debug:
            print("POST stream route: " + url)
        headers = self.s.headers

        headers['Content-Type'] = 'application/json'
        headers['Accept'] ='application/json'
        headers['User-Agent'] = userAgent
        response = self.s.post(url, data=post_json, headers=headers, stream = True)
        self.debug_response(response)
        response.raise_for_status()
        if response.status_code == 204:
            return ""
        return response

    # The Request is streamed, not the Response
    def put_multipart(self, route, fields):
        url = self.host + route
        multipart_data = MultipartEncoder(fields=fields)
        if self.debug:
            print("PUT route: " + url)

        headers = {'Content-Type' : multipart_data.content_type,
                   'Accept' : 'application/json',
#                   'Cache-Control': 'no-cache',
                   'User-Agent':userAgent
                   }
        response = requests.put(url, data=multipart_data, headers=headers,auth=(self.username, self.password))
        self.debug_response(response)
        response.raise_for_status()
        if response.status_code == 204:
            return ""
        try:
            result = response.json()
        except ValueError:
            result = response.text
        return result

    # The Request is streamed, not the Response
    def post_multipart(self, route, fields):
        url = self.host + route
        multipart_data = MultipartEncoder(fields=fields)
        if self.debug:
            print("POST route: " + url)
        headers = {'Content-Type': multipart_data.content_type,
                   #'Accept': 'multipart/form-data', #'application/json',
 #                  'Cache-Control': 'no-cache',
                   'User-Agent': userAgent,
                   }
        response = requests.post(url, data=multipart_data, headers=headers, auth=(self.username, self.password))
        self.debug_response(response)
        response.raise_for_status()
        if response.status_code == 204:
            return ""
        try:
            result = response.json()
        except ValueError:
            result = response.text
        return result

# Network methods

    def save_new_network (self, cx):
        print type(cx)
        if(len(cx) > 0):
            if(cx[len(cx) - 1] is not None):
                if(cx[len(cx) - 1].get('status') is None):
                    # No STATUS element in the array.  Append a new status
                    cx.append({"status" : [ {"error" : "","success" : True} ]})
                else:
                    if(len(cx[len(cx) - 1].get('status')) < 1):
                        # STATUS element found, but the status was empty
                        cx[len(cx) - 1].get('status').append({"error" : "","success" : True})

            if sys.version_info.major == 3:
                stream = io.BytesIO(json.dumps(cx).encode('utf-8'))
            else:
                stream = io.BytesIO(json.dumps(cx, cls=DecimalEncoder))

            return self.save_cx_stream_as_new_network(stream)
        else:
            raise IndexError("Cannot save empty CX.  Please provide a non-empty CX document.")

    # CX Methods
    # Create a network based on a stream from a source CX format
    def save_cx_stream_as_new_network (self, cx_stream):
        ''' Create a new network from a CX stream, optionally providing a provenance history object to be included in the new network.

        :param cx_stream: The network stream.
        :return: The response.
        :rtype: `response object <http://docs.python-requests.org/en/master/user/quickstart/#response-content>`_
        '''
        self.require_auth()
        route = ''
        fields = {}
        if(self.version == "2.0"):
            route = '/network'
            fields = {
                'CXNetworkStream': ('filename', cx_stream, 'application/octet-stream')
            }
        else:
            route = '/network/asCX'
            fields = {
                'CXNetworkStream': ('filename', cx_stream, 'application/octet-stream')
            }

        return self.post_multipart(route, fields)

    # Create a network based on a JSON string or Dict in CX format
    def update_cx_network(self, cx_stream, network_id):
        '''Update the network specified by UUID network_id using the CX stream cx_stream.

        :param cx_stream: The network stream.
        :param network_id: The UUID of the network.
        :type network_id: str
        :return: The response.
        :rtype: `response object <http://docs.python-requests.org/en/master/user/quickstart/#response-content>`_

        '''
        self.require_auth()
        fields = {
            'CXNetworkStream': ('filename', cx_stream, 'application/octet-stream')
        }

        if(self.version == "2.0"):
            route = "/network/%s" % (network_id)
        else:
            route = '/network/asCX/%s' % (network_id)

        return self.put_multipart(route, fields)

    # Get a CX stream for a network
    def get_network_as_cx_stream(self, network_id):
        '''Get the existing network with UUID network_id from the NDEx connection as a CX stream.

        :param network_id: The UUID of the network.
        :type network_id: str
        :return: The response.
        :rtype: `response object <http://docs.python-requests.org/en/master/user/quickstart/#response-content>`_

        '''
        route = ""
        if(self.version == "2.0"):
            route = "/network/%s" % (network_id)
        else:
            route = "/network/%s/asCX" % (network_id)

        return self.get_stream(route)

    # Get a CX stream for a aspect of a network
    def get_network_aspect_as_cx_stream(self, network_id, aspect_name):
        '''Get the specified aspect of the existing network with UUID network_id from the NDEx connection as a CX stream.

        :param network_id: The UUID of the network.
        :param aspect_name: The aspect NAME.
        :type network_id: str
        :return: The response.
        :rtype: `response object <http://docs.python-requests.org/en/master/user/quickstart/#response-content>`_

        '''
        route = ""
        if(self.version == "2.0"):
            route = "/network/%s/aspect/%s" % (network_id, aspect_name)
        else:
            route = "/network/%s/asCX" % (network_id)

        return self.get_stream(route)

    def get_neighborhood_as_cx_stream(self, network_id, search_string, search_depth=1, edge_limit=2500):
        ''' Get a CX stream for a subnetwork of the network specified by UUID network_id and a traversal of search_depth steps around the nodes found by search_string.

        :param network_id: The UUID of the network.
        :type network_id: str
        :param search_string: The search string used to identify the network neighborhood.
        :type search_string: str
        :param search_depth: The depth of the neighborhood from the core nodes identified.
        :type search_depth: int
        :param edge_limit: The maximum size of the neighborhood.
        :type edge_limit: int
        :return: The response.
        :rtype: `response object <http://docs.python-requests.org/en/master/user/quickstart/#response-content>`_

        '''
        if(self.version == "2.0"):
            route = "/search/network/%s/query" % (network_id)
        else:
            route = "/network/%s/query" % (network_id)

        post_data = {'searchString': search_string,
                   'searchDepth': search_depth,
                   'edgeLimit': edge_limit}
        post_json = json.dumps(post_data)
        return self.post_stream(route, post_json=post_json)


    def get_neighborhood(self, network_id, search_string, search_depth=1, edge_limit=2500):
        ''' Get the CX for a subnetwork of the network specified by UUID network_id and a traversal of search_depth steps around the nodes found by search_string.

        :param network_id: The UUID of the network.
        :type network_id: str
        :param search_string: The search string used to identify the network neighborhood.
        :type search_string: str
        :param search_depth: The depth of the neighborhood from the core nodes identified.
        :type search_depth: int
        :param edge_limit: The maximum size of the neighborhood.
        :type edge_limit: int
        :return: The CX json object.
        :rtype: `response object <http://docs.python-requests.org/en/master/user/quickstart/#response-content>`_
        '''
        response = self.get_neighborhood_as_cx_stream(network_id, search_string, search_depth=search_depth, edge_limit=edge_limit)

        if(self.version == "2.0"):
            #response_in_json = response.json()
            #data =  response_in_json["data"]
            #return data
            return response.json()["data"]
        else:
            raise Exception("get_neighborhood is not supported for versions prior to 2.0, use get_neighborhood_as_cx_stream")


    # Search for networks by keywords
    #    network    POST    /network/search/{skipBlocks}/{blockSize}    SimpleNetworkQuery    NetworkSummary[]
    def search_networks(self, search_string="", account_name=None, start=0, size=100, include_groups=False):
        ''' Search for networks based on the search_text, optionally limited to networks owned by the specified account_name.

        :param search_string: The text to search for.
        :type search_string: str
        :param account_name: The account to search
        :type account_name: str
        :param start: The number of blocks to skip. Usually zero, but may be used to page results.
        :type start: int
        :param size: The size of the block.
        :type size: int
        :return: The response.
        :rtype: `response object <http://docs.python-requests.org/en/master/user/quickstart/#response-content>`_

        '''
        post_data = {"searchString" : search_string}
        if self.version == "2.0":
            route = "/search/network?start=%s&size=%s"  % (start, size)
            if include_groups:
                post_data["includeGroups"] = True
        else:
            route = "/network/search/%s/%s" % (start, size)

        if account_name:
            post_data["accountName"] = account_name
        post_json = json.dumps(post_data)
        return self.post(route, post_json)

    def find_networks(self, search_string="", account_name=None, skip_blocks=0, block_size=100):
        print("find_networks is deprecated, please use search_networks")
        return self.search_networks(search_string, account_name, skip_blocks, block_size)

    def search_networks_by_property_filter(self, search_parameter_dict={}, account_name=None, limit=100):
        raise Exception("This function is not supported in NDEx 2.0")
     #   self.require_auth()
     #   route = "/network/searchByProperties"
     #   if account_name:
     #       search_parameter_dict["admin"] = account_name
     #   search_parameter_dict["limit"] = limit
     #   post_json = json.dumps(search_parameter_dict)
     #   return self.post(route, post_json)

    def network_summaries_to_ids(self, network_summaries):
        network_ids = []
        for network in network_summaries:
            network_ids.append(network['externalId'] )
        return network_ids

    #    network    GET    /network/{networkUUID}       NetworkSummary
    def get_network_summary(self, network_id):
        ''' Gets information about a network.

        :param network_id: The UUID of the network.
        :type network_id: str
        :return: The response.
        :rtype: `response object <http://docs.python-requests.org/en/master/user/quickstart/#response-content>`_

        '''
        if(self.version == "2.0"):
            route = "/network/%s/summary" % (network_id)
        else:
            route = "/network/%s" % (network_id)

        return self.get(route)

    def make_network_public(self, network_id):
        ''' Makes the network specified by the network_id public.

        :param network_id: The UUID of the network.
        :type network_id: str
        :return: The response.
        :rtype: `response object <http://docs.python-requests.org/en/master/user/quickstart/#response-content>`_

        '''
        if(self.version == "2.0"):
            return self.set_network_system_properties(network_id, {'visibility': 'PUBLIC'})

        else:
            return self.update_network_profile(network_id, {'visibility': 'PUBLIC'})

    def make_network_private(self, network_id):
        ''' Makes the network specified by the network_id private.

        :param network_id: The UUID of the network.
        :type network_id: str
        :return: The response.
        :rtype: `response object <http://docs.python-requests.org/en/master/user/quickstart/#response-content>`_

        '''
        if(self.version == "2.0"):
            return self.set_network_system_properties(network_id, {'visibility': 'PRIVATE'})

        else:
            return self.update_network_profile(network_id, {'visibility': 'PRIVATE'})


    def get_task_by_id (self, task_id):
        self.require_auth()
        route = "/task/%s" % (task_id)
        return self.get(route)

    def delete_network(self, network_id, retry=5):
        self.require_auth()
        route = "/network/%s" % (network_id)
        count = 0
        while count < retry:
            try:
                return self.delete(route)
            except Exception as inst:
                d = json.loads(inst.response.content)
                if d.get('errorCode').startswith("NDEx_Concurrent_Modification"):
                    print("retry deleting network in 1 second(" + str(count) + ")")
                    count += 1
                    time.sleep(1)
                else:
                    raise inst
        raise Exception("Network is locked after " + retry + " retry.")

    def get_provenance(self, network_id):
        route = "/network/%s/provenance" % (network_id)
        return self.get(route)

    def set_provenance(self, network_id, provenance):
        self.require_auth()
        route = "/network/%s/provenance" % (network_id)
        if isinstance(provenance, dict):
            putJson = json.dumps(provenance)
        else:
            putJson = provenance
        return self.put(route, putJson)


 #   def set_network_flag(self, network_id, parameter, value):
 #       self.require_auth()
 #       route = "/network/%s/setFlag/%s=%s" % (network_id, parameter, value)
 #       return self.get(route)

    def set_read_only(self, network_id, value):
        self.require_auth()
        return self.set_network_system_properties(network_id, {"readOnly": value})

    def set_network_properties(self, network_id, network_properties):
        self.require_auth()
        route = "/network/%s/properties" % (network_id)
        if isinstance(network_properties, list):
            putJson = json.dumps(network_properties)
        elif isinstance(network_properties, basestring):
            putJson = network_properties
        else:
            raise Exception("network_properties must be a string or a list of NdexPropertyValuePair objects")
        return self.put(route, putJson)

    def set_network_system_properties(self, network_id, network_properties):
        self.require_auth()
        route = "/network/%s/systemproperty" % (network_id)
        if isinstance(network_properties, dict):
            putJson = json.dumps(network_properties)
        elif isinstance(network_properties, basestring):
            putJson = network_properties
        else:
            raise Exception("network_properties must be a string or a dict")
        return self.put(route, putJson)

    def update_network_profile(self, network_id, network_profile):
        # network profile attributes that can be updated by this method:
        #   name
        #   description
        #   version

        self.require_auth()
        if isinstance(network_profile, dict):
            if network_profile.get("visibility") and self.version.startswith("2."):
                raise Exception("Ndex 2.x doesn't support setting visibility by this function. Please use make_network_public/private function to set network visibility.")
            json_data = json.dumps(network_profile)
        elif isinstance(network_profile, basestring):
            json_data = network_profile
        else:
            raise Exception("network_profile must be a string or a dict")

        if(self.version == "2.0"):
            route = "/network/%s/profile" % (network_id)
            return self.put(route, json_data)
        else:
            route = "/network/%s/summary" % (network_id)
            return self.post(route, json_data)

    def upload_file(self, filename):
        raise Exception("This function is not supported in this release. Please use the save_new_network function to create new networks in NDEx server.")
 #       self.require_auth()
 #       fields = {

 #           'fileUpload': (filename, open(filename, 'rb'), 'application/octet-stream'),
 #           'filename': os.path.basename(filename)#filename
 #       }
 #       route = '/network/upload'
 #       return self.post_multipart(route, fields)

   # def update_network_membership_by_id(self, account_id, network_id, permission):
   #     route ="/network/{networkId}/member"  % (network_id)
   #     postData = {
   #         "permission": permission,
   #         "networkUUID": network_id,
   #         "userUUID": account_id
   #     }
   #     postJson = json.dumps(postData)
   #     self.post(route, postJson)

    def update_network_group_permission(self, groupid, networkid, permission):
        route = "/network/%s/permission?groupid=%s&permission=%s" % (networkid, groupid, permission)
        self.put(route)

    def update_network_user_permission(self, userid, networkid, permission):
        route = "/network/%s/permission?userid=%s&permission=%s" % (networkid, userid, permission)
        self.put(route)

    # Group methods

    # this group of functions are commented out for this release. We might need to wrap the showcase functions to
    # replace this function
  #  def get_network_summaries_for_group(self, group_name):
  #      route = "/group/network/%s" % (group_name)
  #      return self.get(route)

  #  def get_network_ids_for_group(self, group_name):
  #      return self.network_summaries_to_ids(self.get_network_summaries_for_group(group_name))

    def grant_networks_to_group(self, groupid, networkids, permission="READ"):
        for networkid in networkids:
            self.update_network_group_permission(groupid, networkid, permission)

    # User methods

    def get_user_by_username(self, username):
        route = "/user?username=%s" % (username)
        return self.get(route)

    def get_network_summaries_for_user(self, username):
        network_summaries = self.search_networks("", username, size=1000)

        if (network_summaries and network_summaries['networks']):
            network_summaries_list = network_summaries['networks']
        else:
            network_summaries_list = []

        return network_summaries_list

    def get_network_ids_for_user(self, username):
        network_summaries_list = self.get_network_summaries_for_user(username)

        return self.network_summaries_to_ids(network_summaries_list)

    def grant_network_to_user_by_username(self, username, network_id, permission):
        user = self.get_user_by_username(username).json
        self.update_network_user_permission(user["externalid"], network_id, permission)

    def grant_networks_to_user(self, userid, networkids, permission="READ"):
        for networkid in networkids:
            self.update_network_user_permission(userid, networkid, permission)

    # admin methods

    def update_status(self):
        route = "/admin/status"
        self.status = self.get(route)

    # network set

    def create_networkset (self, name, description):
        route = "/networkset"
        return self.post(route, json.dumps( { "name": name, "description": description}))




class DecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return float(o)
        return super(DecimalEncoder, self).default(o)

